"""
SATOSA microservice that uses a configured ordered list of
attributes that may be asserted by a SAML IdP to construct
a primary identifier or key for the user and assert it as
the value for a configured attribute, for example uid.
"""

import logging
import urllib.parse

import satosa.logging_util as lu
import satosa.micro_services.base
from satosa.response import Redirect


logger = logging.getLogger(__name__)


class PrimaryIdentifier(satosa.micro_services.base.ResponseMicroService):
    """
    Use a configured ordered list of attributes to construct a primary
    identifier for the user and assert it as a particular configured
    attribute. If a primary identifier cannot be found or constructed
    handle the error in a configured way that may be to ignore
    the error or redirect to an external error handling service.
    """
    logprefix = "PRIMARY_IDENTIFIER:"

    def __init__(self, config, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.config = config

    def constructPrimaryIdentifier(self, data, ordered_identifier_candidates):
        """
        Construct and return a primary identifier value from the
        data asserted by the IdP using the ordered list of candidates
        from the configuration.
        """
        logprefix = PrimaryIdentifier.logprefix
        context = self.context

        attributes = data.attributes
        msg = "{} Input attributes {}".format(logprefix, attributes)
        logline = lu.LOG_FMT.format(id=lu.get_session_id(context.state), message=msg)
        logger.debug(logline)

        value = None

        for candidate in ordered_identifier_candidates:
            msg = "{} Considering candidate {}".format(logprefix, candidate)
            logline = lu.LOG_FMT.format(id=lu.get_session_id(context.state), message=msg)
            logger.debug(logline)

            # Get the values asserted by the IdP for the configured list of attribute names for this candidate
            # and substitute None if the IdP did not assert any value for a configured attribute.
            values = [ attributes.get(attribute_name, [None])[0] for attribute_name in candidate['attribute_names'] if attribute_name != 'name_id' ]
            msg = "{} Found candidate values {}".format(logprefix, values)
            logline = lu.LOG_FMT.format(id=lu.get_session_id(context.state), message=msg)
            logger.debug(logline)

            # If one of the configured attribute names is name_id then if there is also a configured
            # name_id_format add the value for the NameID of that format if it was asserted by the IdP
            # or else add the value None.
            if 'name_id' in candidate['attribute_names']:
                candidate_nameid_value = None
                candidate_name_id_format = candidate.get('name_id_format')
                name_id_value = data.subject_id
                name_id_format = data.subject_type
                if (
                    name_id_value
                    and candidate_name_id_format
                    and candidate_name_id_format == name_id_format
                ):
                    msg = "{} IdP asserted NameID {}".format(logprefix, name_id_value)
                    logline = lu.LOG_FMT.format(id=lu.get_session_id(context.state), message=msg)
                    logger.debug(logline)
                    candidate_nameid_value = name_id_value

                # Only add the NameID value asserted by the IdP if it is not already
                # in the list of values. This is necessary because some non-compliant IdPs
                # have been known, for example, to assert the value of eduPersonPrincipalName
                # in the value for SAML2 persistent NameID as well as asserting
                # eduPersonPrincipalName.
                if candidate_nameid_value not in values:
                    msg = "{} Added NameID {} to candidate values".format(
                        logprefix, candidate_nameid_value
                    )
                    logline = lu.LOG_FMT.format(id=lu.get_session_id(context.state), message=msg)
                    logger.debug(logline)
                    values.append(candidate_nameid_value)
                else:
                    msg = "{} NameID {} value also asserted as attribute value".format(
                        logprefix, candidate_nameid_value
                    )
                    logline = logline = lu.LOG_FMT.format(id=lu.get_session_id(context.state), message=msg)
                    logger.warn(logline)

            # If no value was asserted by the IdP for one of the configured list of attribute names
            # for this candidate then go onto the next candidate.
            if None in values:
                msg = "{} Candidate is missing value so skipping".format(logprefix)
                logline = logline = lu.LOG_FMT.format(id=lu.get_session_id(context.state), message=msg)
                logger.debug(logline)
                continue

            # All values for the configured list of attribute names are present
            # so we can create a primary identifer. Add a scope if configured
            # to do so.
            if 'add_scope' in candidate:
                if candidate['add_scope'] == 'issuer_entityid':
                    scope = data.auth_info.issuer
                else:
                    scope = candidate['add_scope']
                msg = "{} Added scope {} to values".format(logprefix, scope)
                logline = lu.LOG_FMT.format(id=lu.get_session_id(context.state), message=msg)
                logger.debug(logline)
                values.append(scope)

            # Concatenate all values to create the primary identifier.
            value = ''.join(values)
            break

        return value

    def process(self, context, data):
        logprefix = PrimaryIdentifier.logprefix
        self.context = context

        # Initialize the configuration to use as the default configuration
        # that is passed during initialization.
        config = self.config

        msg = "{} Using default configuration {}".format(logprefix, config)
        logline = lu.LOG_FMT.format(id=lu.get_session_id(context.state), message=msg)
        logger.debug(logline)

        # Find the entityID for the SP that initiated the flow
        try:
            spEntityID = context.state.state_dict['SATOSA_BASE']['requester']
        except KeyError:
            msg = "{} Unable to determine the entityID for the SP requester".format(logprefix)
            logline = lu.LOG_FMT.format(id=lu.get_session_id(context.state), message=msg)
            logger.error(logline)
            return super().process(context, data)

        msg = "{} entityID for the SP requester is {}".format(logprefix, spEntityID)
        logline = lu.LOG_FMT.format(id=lu.get_session_id(context.state), message=msg)
        logger.debug(logline)

        # Find the entityID for the IdP that issued the assertion
        try:
            idpEntityID = data.auth_info.issuer
        except KeyError:
            msg = "{} Unable to determine the entityID for the IdP issuer".format(logprefix)
            logline = lu.LOG_FMT.format(id=lu.get_session_id(context.state), message=msg)
            logger.error(logline)
            return super().process(context, data)

        # Examine our configuration to determine if there is a per-IdP configuration
        if idpEntityID in self.config:
            config = self.config[idpEntityID]
            msg  = "{} For IdP {} using configuration {}".format(logprefix, idpEntityID, config)
            logline = lu.LOG_FMT.format(id=lu.get_session_id(context.state), message=msg)
            logger.debug(logline)

        # Examine our configuration to determine if there is a per-SP configuration.
        # An SP configuration overrides an IdP configuration when there is a conflict.
        if spEntityID in self.config:
            config = self.config[spEntityID]
            msg = "{} For SP {} using configuration {}".format(logprefix, spEntityID, config)
            logline = lu.LOG_FMT.format(id=lu.get_session_id(context.state), message=msg)
            logger.debug(logline)

        # Obtain configuration details from the per-SP configuration or the default configuration
        try:
            if 'ordered_identifier_candidates' in config:
                ordered_identifier_candidates = config['ordered_identifier_candidates']
            else:
                ordered_identifier_candidates = self.config['ordered_identifier_candidates']
            if 'primary_identifier' in config:
                primary_identifier = config['primary_identifier']
            elif 'primary_identifier' in self.config:
                primary_identifier = self.config['primary_identifier']
            else:
                primary_identifier = 'uid'
            if 'clear_input_attributes' in config:
                clear_input_attributes = config['clear_input_attributes']
            elif 'clear_input_attributes' in self.config:
                clear_input_attributes = self.config['clear_input_attributes']
            else:
                clear_input_attributes = False
            if 'replace_subject_id' in config:
                replace_subject_id = config['replace_subject_id']
            elif 'replace_subject_id' in self.config:
                replace_subject_id = self.config['replace_subject_id']
            else:
                replace_subject_id = False
            if 'ignore' in config:
                ignore = True
            else:
                ignore = False
            if 'on_error' in config:
                on_error = config['on_error']
            elif 'on_error' in self.config:
                on_error = self.config['on_error']
            else:
                on_error = None

        except KeyError as err:
            msg = "{} Configuration '{}' is missing".format(logprefix, err)
            logline = lu.LOG_FMT.format(id=lu.get_session_id(context.state), message=msg)
            logger.error(logline)
            return super().process(context, data)

        # Ignore this SP entirely if so configured.
        if ignore:
            msg = "{} Ignoring SP {}".format(logprefix, spEntityID)
            logline = lu.LOG_FMT.format(id=lu.get_session_id(context.state), message=msg)
            logger.info(logline)
            return super().process(context, data)

        # Construct the primary identifier.
        msg = "{} Constructing primary identifier".format(logprefix)
        logline = lu.LOG_FMT.format(id=lu.get_session_id(context.state), message=msg)
        logger.debug(logline)
        primary_identifier_val = self.constructPrimaryIdentifier(data, ordered_identifier_candidates)

        if not primary_identifier_val:
            msg = "{} No primary identifier found".format(logprefix)
            logline = lu.LOG_FMT.format(id=lu.get_session_id(context.state), message=msg)
            logger.warn(logline)
            if on_error:
                # Redirect to the configured error handling service with
                # the entityIDs for the target SP and IdP used by the user
                # as query string parameters (URL encoded).
                encodedSpEntityID = urllib.parse.quote_plus(spEntityID)
                encodedIdpEntityID = urllib.parse.quote_plus(data.auth_info.issuer)
                url = "{}?sp={}&idp={}".format(on_error, encodedSpEntityID, encodedIdpEntityID)
                msg = "{} Redirecting to {}".format(logprefix, url)
                logline = lu.LOG_FMT.format(id=lu.get_session_id(context.state), message=msg)
                logger.info(logline)
                return Redirect(url)

        msg = "{} Found primary identifier: {}".format(logprefix, primary_identifier_val)
        logline = lu.LOG_FMT.format(id=lu.get_session_id(context.state), message=msg)
        logger.info(logline)

        # Clear input attributes if so configured.
        if clear_input_attributes:
            msg = "{} Clearing values for these input attributes: {}".format(
                logprefix, data.attributes.keys()
            )
            logline = lu.LOG_FMT.format(id=lu.get_session_id(context.state), message=msg)
            logger.debug(logline)
            data.attributes = {}

        if primary_identifier:
            # Set the primary identifier attribute to the value found.
            data.attributes[primary_identifier] = primary_identifier_val
            msg = "{} Setting attribute {} to value {}".format(
                logprefix, primary_identifier, primary_identifier_val
            )
            logline = lu.LOG_FMT.format(id=lu.get_session_id(context.state), message=msg)
            logger.debug(logline)

        # Replace subject_id with the constructed primary identifier if so configured.
        if replace_subject_id:
            msg = "{} Setting subject_id to value {}".format(
                logprefix, primary_identifier_val
            )
            logline = lu.LOG_FMT.format(id=lu.get_session_id(context.state), message=msg)
            logger.debug(logline)
            data.subject_id = primary_identifier_val

        msg = "{} returning data.attributes {}".format(logprefix, str(data.attributes))
        logline = lu.LOG_FMT.format(id=lu.get_session_id(context.state), message=msg)
        logger.debug(logline)
        return super().process(context, data)



from satosa.context import Context
from satosa.internal import InternalData

from .base import RequestMicroService
from ..exception import SATOSAError


class DiscoToTargetIssuerError(SATOSAError):
    """SATOSA exception raised by CustomRouting rules"""


class DiscoToTargetIssuer(RequestMicroService):
    def __init__(self, config:dict, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.disco_endpoints = config['disco_endpoints']
        if not isinstance(self.disco_endpoints, list) or not self.disco_endpoints:
            raise DiscoToTargetIssuerError('disco_endpoints must be a list of str')

    def process(self, context:Context, data:InternalData):
        context.state[self.name] = {
            'target_frontend': context.target_frontend,
            'internal_data': data.to_dict(),
        }
        return super().process(context, data)

    def register_endpoints(self):
        """
        URL mapping of additional endpoints this micro service needs to register for callbacks.

        Example of a mapping from the url path '/callback' to the callback() method of a micro service:
            reg_endp = [
                ('^/callback1$', self.callback),
            ]

        :rtype List[Tuple[str, Callable[[satosa.context.Context, Any], satosa.response.Response]]]

        :return: A list with functions and args bound to a specific endpoint url,
                 [(regexp, Callable[[satosa.context.Context], satosa.response.Response]), ...]
        """

        return [
            (path , self._handle_disco_response)
            for path in self.disco_endpoints
        ]

    def _handle_disco_response(self, context:Context):
        target_issuer = context.request.get('entityID')
        if not target_issuer:
            raise DiscoToTargetIssuerError('no valid entity_id in the disco response')

        target_frontend = context.state.get(self.name, {}).get('target_frontend')
        data_serialized = context.state.get(self.name, {}).get('internal_data', {})
        data = InternalData.from_dict(data_serialized)

        context.target_frontend = target_frontend
        context.decorate(Context.KEY_TARGET_ENTITYID, target_issuer)
        return super().process(context, data)



import argparse
import os
import sys

from werkzeug.serving import run_simple

from satosa.proxy_server import make_app
from satosa.satosa_config import SATOSAConfig

config_file = os.environ.get("SATOSA_CONFIG", "proxy_conf.yaml")
satosa_config = SATOSAConfig(config_file)
app = make_app(satosa_config)


def main():
    global app

    parser = argparse.ArgumentParser(description="Process some integers.")
    parser.add_argument("port", type=int)
    parser.add_argument("--keyfile", type=str)
    parser.add_argument("--certfile", type=str)
    parser.add_argument("--host", type=str)
    args = parser.parse_args()

    if (args.keyfile and not args.certfile) or (args.certfile and not args.keyfile):
        print("Both keyfile and certfile must be specified for HTTPS.")
        sys.exit(1)

    ssl_context = (
        (args.certfile, args.keyfile)
        if args.keyfile and args.certfile
        else None
    )
    host = args.host or "localhost"
    run_simple(host, args.port, app, ssl_context=ssl_context)


if __name__ == "__main__":
    main()



"""
Some help functions to load satosa backend and frontend modules
"""
import json
import logging
import sys
from contextlib import contextmanager
from pydoc import locate

from satosa.yaml import load as yaml_load
from satosa.yaml import YAMLError

from .backends.base import BackendModule
from .exception import SATOSAConfigurationError
from .frontends.base import FrontendModule
from .micro_services.base import (MicroService, RequestMicroService, ResponseMicroService)

logger = logging.getLogger(__name__)


@contextmanager
def prepend_to_import_path(import_paths):
    import_paths = import_paths or []
    for p in reversed(import_paths):  # insert the specified plugin paths in the same order
        sys.path.insert(0, p)
    yield
    del sys.path[0:len(import_paths)]  # restore sys.path


def load_backends(config, callback, internal_attributes):
    """
    Load all backend modules specified in the config

    :type config: satosa.satosa_config.SATOSAConfig
    :type callback:
    (satosa.context.Context, satosa.internal.InternalData) -> satosa.response.Response
    :type internal_attributes: dict[string, dict[str, str | list[str]]]
    :rtype: Sequence[satosa.backends.base.BackendModule]

    :param config: The configuration of the satosa proxy
    :param callback: Function that will be called by the backend after the authentication is done.
    :return: A list of backend modules
    """
    backend_modules = _load_plugins(
        config.get("CUSTOM_PLUGIN_MODULE_PATHS"),
        config["BACKEND_MODULES"],
        backend_filter, config["BASE"],
        internal_attributes, callback)
    logger.info("Setup backends: {}".format([backend.name for backend in backend_modules]))
    return backend_modules


def load_frontends(config, callback, internal_attributes):
    """
    Load all frontend modules specified in the config

    :type config: satosa.satosa_config.SATOSAConfig
    :type callback:
    (satosa.context.Context, satosa.internal.InternalData) -> satosa.response.Response
    :type internal_attributes: dict[string, dict[str, str | list[str]]]
    :rtype: Sequence[satosa.frontends.base.FrontendModule]

    :param config: The configuration of the satosa proxy
    :param callback: Function that will be called by the frontend after the authentication request
    has been processed.
    :return: A list of frontend modules
    """
    frontend_modules = _load_plugins(config.get("CUSTOM_PLUGIN_MODULE_PATHS"), config["FRONTEND_MODULES"],
                                     frontend_filter, config["BASE"], internal_attributes, callback)
    logger.info("Setup frontends: {}".format([frontend.name for frontend in frontend_modules]))
    return frontend_modules


def backend_filter(cls):
    """
    Verify that the type proper subclass of BackendModule.

    :type cls: type
    :rtype: bool

    :param cls: A class object
    :return: True if match, else false
    """
    return issubclass(cls, BackendModule) and cls != BackendModule


def frontend_filter(cls):
    """
    Verify that the type proper subclass of FrontendModule.

    :type cls: type
    :rtype: bool

    :param cls: A class object
    :return: True if match, else false
    """
    return issubclass(cls, FrontendModule) and cls != FrontendModule


def _micro_service_filter(cls):
    """
    Will only give a find on classes that is a subclass of MicroService, with the exception that
    the class is not allowed to be a direct ResponseMicroService or RequestMicroService.

    :type cls: type
    :rtype: bool

    :param cls: A class object
    :return: True if match, else false
    """
    is_microservice_module = issubclass(cls, MicroService)
    is_correct_subclass = cls != MicroService and cls != ResponseMicroService and cls != RequestMicroService
    return is_microservice_module and is_correct_subclass


def _request_micro_service_filter(cls):
    """
    Will only give a find on classes that is a subclass of RequestMicroService.
    Use this filter to only find frontend plugins

    :type cls: type
    :rtype: bool

    :param cls: A class object
    :return: True if match, else false
    """
    return issubclass(cls, RequestMicroService) and cls != RequestMicroService


def _response_micro_service_filter(cls):
    """
    Will only give a find on classes that is a subclass of ResponseMicroService.
    Use this filter to only find frontend plugins

    :type cls: type
    :rtype: bool

    :param cls: A class object
    :return: True if match, else false
    """
    return issubclass(cls, ResponseMicroService) and cls != ResponseMicroService


def _load_plugin_config(config):
    try:
        return yaml_load(config)
    except YAMLError as exc:
        if hasattr(exc, 'problem_mark'):
            mark = exc.problem_mark
            logger.error("Error position: ({line}:{column})".format(line=mark.line + 1, column=mark.column + 1))
            raise SATOSAConfigurationError("The configuration is corrupt.") from exc


def _load_plugins(plugin_paths, plugins, plugin_filter, base_url, internal_attributes, callback):
    """
    Loads endpoint plugins

    :type plugin_paths: list[str]
    :type plugins: list[str]
    :type plugin_filter: (type | str) -> bool
    :type internal_attributes: dict[string, dict[str, str | list[str]]]
    :rtype list[satosa.plugin_base.endpoint.InterfaceModulePlugin]

    :param plugin_paths: Path to the plugin directory
    :param plugins: A list with the name of the plugin files
    :param plugin_filter: Filter what to load from the module file
    :param args: Arguments to the plugin
    :return: A list with all the loaded plugins
    """
    loaded_plugin_modules = []
    with prepend_to_import_path(plugin_paths):
        for plugin_config in plugins:
            try:
                module_class = _load_endpoint_module(plugin_config, plugin_filter)
            except SATOSAConfigurationError as e:
                raise SATOSAConfigurationError("Configuration error in {}".format(json.dumps(plugin_config))) from e

            if module_class:
                module_config = _replace_variables_in_plugin_module_config(plugin_config["config"], base_url,
                                                                           plugin_config["name"])
                instance = module_class(callback, internal_attributes, module_config, base_url,
                                        plugin_config["name"])
                loaded_plugin_modules.append(instance)
    return loaded_plugin_modules


def _load_endpoint_module(plugin_config, plugin_filter):
    _mandatory_params = ("name", "module", "config")
    if not all(k in plugin_config for k in _mandatory_params):
        raise SATOSAConfigurationError(
            "Missing mandatory plugin configuration parameter: {}".format(_mandatory_params))

    return _load_plugin_module(plugin_config, plugin_filter)


def _load_plugin_module(plugin_config, plugin_filter):
    module_class = locate(plugin_config["module"])
    if not module_class:
        raise ValueError("Can't find module '%s'" % plugin_config["module"])
    if not plugin_filter(module_class):
        return None

    return module_class


def _load_microservice(plugin_config, plugin_filter):
    _mandatory_params = ("name", "module")
    if not all(k in plugin_config for k in _mandatory_params):
        raise SATOSAConfigurationError(
            "Missing mandatory plugin configuration parameter: {}".format(_mandatory_params))

    return _load_plugin_module(plugin_config, plugin_filter)


def _load_microservices(plugin_paths, plugins, plugin_filter, internal_attributes, base_url):
    loaded_plugin_modules = []
    with prepend_to_import_path(plugin_paths):
        for plugin_config in plugins:
            try:
                module_class = _load_microservice(plugin_config, plugin_filter)
            except SATOSAConfigurationError as e:
                raise SATOSAConfigurationError("Configuration error in {}".format(json.dumps(plugin_config))) from e

            if module_class:
                instance = module_class(internal_attributes=internal_attributes, config=plugin_config.get("config"),
                                        name=plugin_config["name"], base_url=base_url)
                loaded_plugin_modules.append(instance)

    return loaded_plugin_modules


def _replace_variables_in_plugin_module_config(module_config, base_url, name):
    config = json.dumps(module_config)
    replace = [
        ("<base_url>", base_url),
        ("<name>", name)
    ]
    for _replace in replace:
        config = config.replace(_replace[0], _replace[1])
    return json.loads(config)


def load_request_microservices(plugin_path, plugins, internal_attributes, base_url):
    """
    Loads request micro services (handling incoming requests).

    :type plugin_path: list[str]
    :type plugins: list[str]
    :type internal_attributes: dict[string, dict[str, str | list[str]]]
    :type base_url: str
    :rtype satosa.micro_service.service_base.RequestMicroService

    :param plugin_path: Path to the plugin directory
    :param plugins: A list with the name of the plugin files
    :param: base_url: base url of the SATOSA server
    :return: Request micro service
    """
    request_services = _load_microservices(plugin_path, plugins, _request_micro_service_filter, internal_attributes,
                                           base_url)
    logger.info("Loaded request micro services: {}".format([type(k).__name__ for k in request_services]))
    return request_services


def load_response_microservices(plugin_path, plugins, internal_attributes, base_url):
    """
    Loads response micro services (handling outgoing responses).

    :type plugin_path: list[str]
    :type plugins: list[str]
    :type internal_attributes: dict[string, dict[str, str | list[str]]]
    :type base_url: str
    :rtype satosa.micro_service.service_base.ResponseMicroService

    :param plugin_path: Path to the plugin directory
    :param plugins: A list with the name of the plugin files
    :param: base_url: base url of the SATOSA server
    :return: Response micro service
    """
    response_services = _load_microservices(plugin_path, plugins, _response_micro_service_filter, internal_attributes,
                                            base_url)
    logger.info("Loaded response micro services:{}".format([type(k).__name__ for k in response_services]))
    return response_services



"""
A OpenID Connect frontend module for the satosa proxy
"""

import json
import logging
from collections import defaultdict
from urllib.parse import urlencode, urlparse

from jwkest.jwk import rsa_load, RSAKey

from oic.oic import scope2claims
from oic.oic.message import AuthorizationRequest
from oic.oic.message import AuthorizationErrorResponse
from oic.oic.message import TokenErrorResponse
from oic.oic.message import UserInfoErrorResponse
from oic.oic.provider import RegistrationEndpoint
from oic.oic.provider import AuthorizationEndpoint
from oic.oic.provider import TokenEndpoint
from oic.oic.provider import UserinfoEndpoint

from pyop.access_token import AccessToken
from pyop.authz_state import AuthorizationState
from pyop.exceptions import InvalidAuthenticationRequest
from pyop.exceptions import InvalidClientRegistrationRequest
from pyop.exceptions import InvalidClientAuthentication
from pyop.exceptions import OAuthError
from pyop.exceptions import BearerTokenError
from pyop.exceptions import InvalidAccessToken
from pyop.provider import Provider
from pyop.storage import StorageBase
from pyop.subject_identifier import HashBasedSubjectIdentifierFactory
from pyop.userinfo import Userinfo
from pyop.util import should_fragment_encode

from .base import FrontendModule
from ..response import BadRequest, Created
from ..response import SeeOther, Response
from ..response import Unauthorized
from ..util import rndstr

import satosa.logging_util as lu
from satosa.internal import InternalData


logger = logging.getLogger(__name__)


class MirrorPublicSubjectIdentifierFactory(HashBasedSubjectIdentifierFactory):
    def create_public_identifier(self, user_id):
        return user_id


class OpenIDConnectFrontend(FrontendModule):
    """
    A OpenID Connect frontend module
    """

    def __init__(self, auth_req_callback_func, internal_attributes, conf, base_url, name):
        _validate_config(conf)
        super().__init__(auth_req_callback_func, internal_attributes, base_url, name)

        self.config = conf
        provider_config = self.config["provider"]
        provider_config["issuer"] = base_url

        self.signing_key = RSAKey(
            key=rsa_load(self.config["signing_key_path"]),
            use="sig",
            alg="RS256",
            kid=self.config.get("signing_key_id", ""),
        )

        db_uri = self.config.get("db_uri")
        self.stateless = db_uri and StorageBase.type(db_uri) == "stateless"
        self.user_db = (
            StorageBase.from_uri(db_uri, db_name="satosa", collection="authz_codes")
            if db_uri and not self.stateless
            else {}
        )

        sub_hash_salt = self.config.get("sub_hash_salt", rndstr(16))
        mirror_public = self.config.get("sub_mirror_public", False)
        authz_state = _init_authorization_state(
            provider_config, db_uri, sub_hash_salt, mirror_public
        )

        client_db_uri = self.config.get("client_db_uri")
        cdb_file = self.config.get("client_db_path")
        if client_db_uri:
            cdb = StorageBase.from_uri(
                client_db_uri, db_name="satosa", collection="clients", ttl=None
            )
        elif cdb_file:
            with open(cdb_file) as f:
                cdb = json.loads(f.read())
        else:
            cdb = {}

        self.endpoint_baseurl = "{}/{}".format(self.base_url, self.name)
        self.provider = _create_provider(
            provider_config,
            self.endpoint_baseurl,
            self.internal_attributes,
            self.signing_key,
            authz_state,
            self.user_db,
            cdb,
        )

    def _get_extra_id_token_claims(self, user_id, client_id):
        if "extra_id_token_claims" in self.config["provider"]:
            config = self.config["provider"]["extra_id_token_claims"].get(client_id, [])
            if type(config) is list and len(config) > 0:
                requested_claims = {k: None for k in config}
                return self.provider.userinfo.get_claims_for(user_id, requested_claims)
        return {}

    def handle_authn_response(self, context, internal_resp):
        """
        See super class method satosa.frontends.base.FrontendModule#handle_authn_response
        :type context: satosa.context.Context
        :type internal_response: satosa.internal.InternalData
        :rtype oic.utils.http_util.Response
        """

        auth_req = self._get_authn_request_from_state(context.state)

        claims = self.converter.from_internal("openid", internal_resp.attributes)
        # Filter unset claims
        claims = {k: v for k, v in claims.items() if v}
        self.user_db[internal_resp.subject_id] = dict(
            combine_claim_values(claims.items())
        )
        auth_resp = self.provider.authorize(
            auth_req,
            internal_resp.subject_id,
            extra_id_token_claims=lambda user_id, client_id:
                self._get_extra_id_token_claims(user_id, client_id),
        )

        if self.stateless:
            del self.user_db[internal_resp.subject_id]

        del context.state[self.name]
        http_response = auth_resp.request(auth_req["redirect_uri"], should_fragment_encode(auth_req))
        return SeeOther(http_response)

    def handle_backend_error(self, exception):
        """
        See super class satosa.frontends.base.FrontendModule
        :type exception: satosa.exception.SATOSAError
        :rtype: oic.utils.http_util.Response
        """
        auth_req = self._get_authn_request_from_state(exception.state)
        # If the client sent us a state parameter, we should reflect it back according to the spec
        if 'state' in auth_req:
            error_resp = AuthorizationErrorResponse(error="access_denied",
                                                    error_description=exception.message,
                                                    state=auth_req['state'])
        else:
            error_resp = AuthorizationErrorResponse(error="access_denied",
                                                    error_description=exception.message)
        msg = exception.message
        logline = lu.LOG_FMT.format(id=lu.get_session_id(exception.state), message=msg)
        logger.debug(logline)
        return SeeOther(error_resp.request(auth_req["redirect_uri"], should_fragment_encode(auth_req)))

    def register_endpoints(self, backend_names):
        """
        See super class satosa.frontends.base.FrontendModule
        :type backend_names: list[str]
        :rtype: list[(str, ((satosa.context.Context, Any) -> satosa.response.Response, Any))]
        :raise ValueError: if more than one backend is configured
        """
        backend_name = None
        if len(backend_names) != 1:
            # only supports one backend since there currently is no way to publish multiple authorization endpoints
            # in configuration information and there is no other standard way of authorization_endpoint discovery
            # similar to SAML entity discovery
            # this can be circumvented with a custom RequestMicroService which handles the routing based on something
            # in the authentication request
            logline = (
                "More than one backend is configured, "
                "make sure to provide a custom routing micro service "
                "to determine which backend should be used per request."
            )
            logger.warning(logline)
        else:
            backend_name = backend_names[0]

        provider_config = ("^.well-known/openid-configuration$", self.provider_config)
        jwks_uri = ("^{}/jwks$".format(self.name), self.jwks)

        if backend_name:
            # if there is only one backend, include its name in the path so the default routing can work
            auth_endpoint = "{}/{}/{}/{}".format(self.base_url, backend_name, self.name, AuthorizationEndpoint.url)
            self.provider.configuration_information["authorization_endpoint"] = auth_endpoint
            auth_path = urlparse(auth_endpoint).path.lstrip("/")
        else:
            auth_path = "{}/{}".format(self.name, AuthorizationEndpoint.url)

        authentication = ("^{}$".format(auth_path), self.handle_authn_request)
        url_map = [provider_config, jwks_uri, authentication]

        if any("code" in v for v in self.provider.configuration_information["response_types_supported"]):
            self.provider.configuration_information["token_endpoint"] = "{}/{}".format(
                self.endpoint_baseurl, TokenEndpoint.url
            )
            token_endpoint = (
                "^{}/{}".format(self.name, TokenEndpoint.url), self.token_endpoint
            )
            url_map.append(token_endpoint)

            self.provider.configuration_information["userinfo_endpoint"] = (
                "{}/{}".format(self.endpoint_baseurl, UserinfoEndpoint.url)
            )
            userinfo_endpoint = (
                "^{}/{}".format(self.name, UserinfoEndpoint.url), self.userinfo_endpoint
            )
            url_map.append(userinfo_endpoint)

        if "registration_endpoint" in self.provider.configuration_information:
            client_registration = (
                "^{}/{}".format(self.name, RegistrationEndpoint.url),
                self.client_registration,
            )
            url_map.append(client_registration)

        return url_map

    def _get_authn_request_from_state(self, state):
        """
        Extract the clietns request stoed in the SATOSA state.
        :type state: satosa.state.State
        :rtype: oic.oic.message.AuthorizationRequest

        :param state: the current state
        :return: the parsed authentication request
        """
        return AuthorizationRequest().deserialize(state[self.name]["oidc_request"])

    def client_registration(self, context):
        """
        Handle the OIDC dynamic client registration.
        :type context: satosa.context.Context
        :rtype: oic.utils.http_util.Response

        :param context: the current context
        :return: HTTP response to the client
        """
        try:
            resp = self.provider.handle_client_registration_request(json.dumps(context.request))
            return Created(resp.to_json(), content="application/json")
        except InvalidClientRegistrationRequest as e:
            return BadRequest(e.to_json(), content="application/json")

    def provider_config(self, context):
        """
        Construct the provider configuration information (served at /.well-known/openid-configuration).
        :type context: satosa.context.Context
        :rtype: oic.utils.http_util.Response

        :param context: the current context
        :return: HTTP response to the client
        """
        return Response(self.provider.provider_configuration.to_json(), content="application/json")

    def _get_approved_attributes(self, provider_supported_claims, authn_req):
        requested_claims = list(
            scope2claims(
                authn_req["scope"], self.config["provider"].get("extra_scopes")
            ).keys()
        )
        if "claims" in authn_req:
            for k in ["id_token", "userinfo"]:
                if k in authn_req["claims"]:
                    requested_claims.extend(authn_req["claims"][k].keys())
        return set(provider_supported_claims).intersection(set(requested_claims))

    def _handle_authn_request(self, context):
        """
        Parse and verify the authentication request into an internal request.
        :type context: satosa.context.Context
        :rtype: satosa.internal.InternalData

        :param context: the current context
        :return: the internal request
        """
        request = urlencode(context.request)
        msg = "Authn req from client: {}".format(request)
        logline = lu.LOG_FMT.format(id=lu.get_session_id(context.state), message=msg)
        logger.debug(logline)

        try:
            authn_req = self.provider.parse_authentication_request(request)
        except InvalidAuthenticationRequest as e:
            msg = "Error in authn req: {}".format(str(e))
            logline = lu.LOG_FMT.format(id=lu.get_session_id(context.state), message=msg)
            logger.error(logline)
            error_url = e.to_error_url()

            if error_url:
                return SeeOther(error_url)
            else:
                return BadRequest("Something went wrong: {}".format(str(e)))

        client_id = authn_req["client_id"]
        context.state[self.name] = {"oidc_request": request}
        subject_type = self.provider.clients[client_id].get("subject_type", "pairwise")
        client_name = self.provider.clients[client_id].get("client_name")
        if client_name:
            # TODO should process client names for all languages, see OIDC Registration, Section 2.1
            requester_name = [{"lang": "en", "text": client_name}]
        else:
            requester_name = None
        internal_req = InternalData(
            subject_type=subject_type,
            requester=client_id,
            requester_name=requester_name,
        )

        internal_req.attributes = self.converter.to_internal_filter(
            "openid", self._get_approved_attributes(self.provider.configuration_information["claims_supported"],
                                                    authn_req))
        return internal_req

    def handle_authn_request(self, context):
        """
        Handle an authentication request and pass it on to the backend.
        :type context: satosa.context.Context
        :rtype: oic.utils.http_util.Response

        :param context: the current context
        :return: HTTP response to the client
        """
        internal_req = self._handle_authn_request(context)
        if not isinstance(internal_req, InternalData):
            return internal_req
        return self.auth_req_callback_func(context, internal_req)

    def jwks(self, context):
        """
        Construct the JWKS document (served at /jwks).
        :type context: satosa.context.Context
        :rtype: oic.utils.http_util.Response

        :param context: the current context
        :return: HTTP response to the client
        """
        return Response(json.dumps(self.provider.jwks), content="application/json")

    def token_endpoint(self, context):
        """
        Handle token requests (served at /token).
        :type context: satosa.context.Context
        :rtype: oic.utils.http_util.Response

        :param context: the current context
        :return: HTTP response to the client
        """
        headers = {"Authorization": context.request_authorization}
        try:
            response = self.provider.handle_token_request(
                urlencode(context.request),
                headers,
                lambda user_id, client_id: self._get_extra_id_token_claims(user_id, client_id))
            return Response(response.to_json(), content="application/json")
        except InvalidClientAuthentication as e:
            logline = "invalid client authentication at token endpoint"
            logger.debug(logline, exc_info=True)
            error_resp = TokenErrorResponse(error='invalid_client', error_description=str(e))
            response = Unauthorized(error_resp.to_json(), headers=[("WWW-Authenticate", "Basic")],
                                    content="application/json")
            return response
        except OAuthError as e:
            logline = "invalid request: {}".format(str(e))
            logger.debug(logline, exc_info=True)
            error_resp = TokenErrorResponse(error=e.oauth_error, error_description=str(e))
            return BadRequest(error_resp.to_json(), content="application/json")

    def userinfo_endpoint(self, context):
        headers = {"Authorization": context.request_authorization}

        try:
            response = self.provider.handle_userinfo_request(
                request=urlencode(context.request),
                http_headers=headers,
            )
            return Response(response.to_json(), content="application/json")
        except (BearerTokenError, InvalidAccessToken) as e:
            error_resp = UserInfoErrorResponse(error='invalid_token', error_description=str(e))
            response = Unauthorized(error_resp.to_json(), headers=[("WWW-Authenticate", AccessToken.BEARER_TOKEN_TYPE)],
                                    content="application/json")
            return response


def _validate_config(config):
    """
    Validates that all necessary config parameters are specified.
    :type config: dict[str, dict[str, Any] | str]
    :param config: the module config
    """
    if config is None:
        raise ValueError("OIDCFrontend configuration can't be 'None'.")

    for k in {"signing_key_path", "provider"}:
        if k not in config:
            raise ValueError("Missing configuration parameter '{}' for OpenID Connect frontend.".format(k))

    if "signing_key_id" in config and type(config["signing_key_id"]) is not str:
        raise ValueError(
            "The configuration parameter 'signing_key_id' is not defined as a string for OpenID Connect frontend.")


def _create_provider(
    provider_config,
    endpoint_baseurl,
    internal_attributes,
    signing_key,
    authz_state,
    user_db,
    cdb,
):
    response_types_supported = provider_config.get("response_types_supported", ["id_token"])
    subject_types_supported = provider_config.get("subject_types_supported", ["pairwise"])
    scopes_supported = provider_config.get("scopes_supported", ["openid"])
    extra_scopes = provider_config.get("extra_scopes")
    capabilities = {
        "issuer": provider_config["issuer"],
        "authorization_endpoint": "{}/{}".format(endpoint_baseurl, AuthorizationEndpoint.url),
        "jwks_uri": "{}/jwks".format(endpoint_baseurl),
        "response_types_supported": response_types_supported,
        "id_token_signing_alg_values_supported": [signing_key.alg],
        "response_modes_supported": ["fragment", "query"],
        "subject_types_supported": subject_types_supported,
        "claim_types_supported": ["normal"],
        "claims_parameter_supported": True,
        "claims_supported": [
            attribute_map["openid"][0]
            for attribute_map in internal_attributes["attributes"].values()
            if "openid" in attribute_map
        ],
        "request_parameter_supported": False,
        "request_uri_parameter_supported": False,
        "scopes_supported": scopes_supported
    }

    if 'code' in response_types_supported:
        capabilities["token_endpoint"] = "{}/{}".format(
            endpoint_baseurl, TokenEndpoint.url
        )

    if provider_config.get("client_registration_supported", False):
        capabilities["registration_endpoint"] = "{}/{}".format(
            endpoint_baseurl, RegistrationEndpoint.url
        )

    provider = Provider(
        signing_key,
        capabilities,
        authz_state,
        cdb,
        Userinfo(user_db),
        extra_scopes=extra_scopes,
        id_token_lifetime=provider_config.get("id_token_lifetime", 3600),
    )
    return provider


def _init_authorization_state(
    provider_config, db_uri, sub_hash_salt, mirror_public=False
):
    if db_uri:
        authz_code_db = StorageBase.from_uri(
            db_uri,
            db_name="satosa",
            collection="authz_codes",
            ttl=provider_config.get("authorization_code_lifetime", 600),
        )
        access_token_db = StorageBase.from_uri(
            db_uri,
            db_name="satosa",
            collection="access_tokens",
            ttl=provider_config.get("access_token_lifetime", 3600),
        )
        refresh_token_db = StorageBase.from_uri(
            db_uri,
            db_name="satosa",
            collection="refresh_tokens",
            ttl=provider_config.get("refresh_token_lifetime", None),
        )
        sub_db = StorageBase.from_uri(
            db_uri, db_name="satosa", collection="subject_identifiers", ttl=None
        )
    else:
        authz_code_db = None
        access_token_db = None
        refresh_token_db = None
        sub_db = None

    token_lifetimes = {
        k: provider_config[k]
        for k in [
            "authorization_code_lifetime",
            "access_token_lifetime",
            "refresh_token_lifetime",
            "refresh_token_threshold",
        ]
        if k in provider_config
    }

    subject_id_factory = (
        MirrorPublicSubjectIdentifierFactory(sub_hash_salt)
        if mirror_public
        else HashBasedSubjectIdentifierFactory(sub_hash_salt)
    )
    return AuthorizationState(
        subject_id_factory,
        authz_code_db,
        access_token_db,
        refresh_token_db,
        sub_db,
        **token_lifetimes,
    )


def combine_return_input(values):
    return values


def combine_select_first_value(values):
    return values[0]


def combine_join_by_space(values):
    return " ".join(values)


combine_values_by_claim = defaultdict(
    lambda: combine_return_input,
    {
        "sub": combine_select_first_value,
        "name": combine_select_first_value,
        "given_name": combine_join_by_space,
        "family_name": combine_join_by_space,
        "middle_name": combine_join_by_space,
        "nickname": combine_select_first_value,
        "preferred_username": combine_select_first_value,
        "profile": combine_select_first_value,
        "picture": combine_select_first_value,
        "website": combine_select_first_value,
        "email": combine_select_first_value,
        "email_verified": combine_select_first_value,
        "gender": combine_select_first_value,
        "birthdate": combine_select_first_value,
        "zoneinfo": combine_select_first_value,
        "locale": combine_select_first_value,
        "phone_number": combine_select_first_value,
        "phone_number_verified": combine_select_first_value,
        "address": combine_select_first_value,
        "updated_at": combine_select_first_value,
    },
)


def combine_claim_values(claim_items):
    claims = (
        (name, combine_values_by_claim[name](values))
        for name, values in claim_items
    )
    return claims



"""
OIDC/OAuth2 backend module.
"""
import datetime
import logging
from urllib.parse import urlparse

from idpyoidc.client.oauth2.stand_alone_client import StandAloneClient
from idpyoidc.server.user_authn.authn_context import UNSPECIFIED

from satosa.backends.base import BackendModule
from satosa.internal import AuthenticationInformation
from satosa.internal import InternalData
import satosa.logging_util as lu
from ..exception import SATOSAAuthenticationError
from ..exception import SATOSAError
from ..response import Redirect


UTC = datetime.timezone.utc
logger = logging.getLogger(__name__)


class IdpyOIDCBackend(BackendModule):
    """
    Backend module for OIDC and OAuth 2.0, can be directly used.
    """

    def __init__(self, auth_callback_func, internal_attributes, config, base_url, name):
        """
        OIDC backend module.
        :param auth_callback_func: Callback should be called by the module after the authorization
        in the backend is done.
        :param internal_attributes: Mapping dictionary between SATOSA internal attribute names and
        the names returned by underlying IdP's/OP's as well as what attributes the calling SP's and
        RP's expects namevice.
        :param config: Configuration parameters for the module.
        :param base_url: base url of the service
        :param name: name of the plugin

        :type auth_callback_func:
        (satosa.context.Context, satosa.internal.InternalData) -> satosa.response.Response
        :type internal_attributes: dict[string, dict[str, str | list[str]]]
        :type config: dict[str, dict[str, str] | list[str]]
        :type base_url: str
        :type name: str
        """
        super().__init__(auth_callback_func, internal_attributes, base_url, name)
        # self.auth_callback_func = auth_callback_func
        # self.config = config
        self.client = StandAloneClient(config=config["client"], client_type="oidc")
        self.client.do_provider_info()
        self.client.do_client_registration()

        _redirect_uris = self.client.context.claims.get_usage('redirect_uris')
        if not _redirect_uris:
            raise SATOSAError("Missing path in redirect uri")
        self.redirect_path = urlparse(_redirect_uris[0]).path

    def start_auth(self, context, internal_request):
        """
        See super class method satosa.backends.base#start_auth

        :type context: satosa.context.Context
        :type internal_request: satosa.internal.InternalData
        :rtype satosa.response.Redirect
        """
        login_url = self.client.init_authorization()
        return Redirect(login_url)

    def register_endpoints(self):
        """
        Creates a list of all the endpoints this backend module needs to listen to. In this case
        it's the authentication response from the underlying OP that is redirected from the OP to
        the proxy.
        :rtype: Sequence[(str, Callable[[satosa.context.Context], satosa.response.Response]]
        :return: A list that can be used to map the request to SATOSA to this endpoint.
        """
        url_map = []
        url_map.append((f"^{self.redirect_path.lstrip('/')}$", self.response_endpoint))
        return url_map

    def response_endpoint(self, context, *args):
        """
        Handles the authentication response from the OP.
        :type context: satosa.context.Context
        :type args: Any
        :rtype: satosa.response.Response

        :param context: SATOSA context
        :param args: None
        :return:
        """

        _info = self.client.finalize(context.request)
        self._check_error_response(_info, context)
        userinfo = _info.get('userinfo')
        id_token = _info.get('id_token')

        if not id_token and not userinfo:
            msg = "No id_token or userinfo, nothing to do.."
            logline = lu.LOG_FMT.format(id=lu.get_session_id(context.state), message=msg)
            logger.error(logline)
            raise SATOSAAuthenticationError(context.state, "No user info available.")

        all_user_claims = dict(list(userinfo.items()) + list(id_token.items()))
        msg = "UserInfo: {}".format(all_user_claims)
        logline = lu.LOG_FMT.format(id=lu.get_session_id(context.state), message=msg)
        logger.debug(logline)
        internal_resp = self._translate_response(all_user_claims, _info["issuer"])
        return self.auth_callback_func(context, internal_resp)

    def _translate_response(self, response, issuer):
        """
        Translates oidc response to SATOSA internal response.
        :type response: dict[str, str]
        :type issuer: str
        :type subject_type: str
        :rtype: InternalData

        :param response: Dictioary with attribute name as key.
        :param issuer: The oidc op that gave the repsonse.
        :param subject_type: public or pairwise according to oidc standard.
        :return: A SATOSA internal response.
        """
        timestamp_epoch = (
            response.get("auth_time")
            or response.get("iat")
            or int(datetime.datetime.now(UTC).timestamp())
        )
        timestamp_dt = datetime.datetime.fromtimestamp(timestamp_epoch, UTC)
        timestamp_iso = timestamp_dt.isoformat().replace("+00:00", "Z")
        auth_class_ref = response.get("acr") or response.get("amr") or UNSPECIFIED
        auth_info = AuthenticationInformation(auth_class_ref, timestamp_iso, issuer)

        internal_resp = InternalData(auth_info=auth_info)
        internal_resp.attributes = self.converter.to_internal("openid", response)
        internal_resp.subject_id = response["sub"]
        return internal_resp

    def _check_error_response(self, response, context):
        """
        Check if the response is an error response.
        :param response: the response from finalize()
        :type response: oic.oic.message
        :raise SATOSAAuthenticationError: if the response is an OAuth error response
        """
        if "error" in response:
            msg = "{name} error: {error} {description}".format(
                name=type(response).__name__,
                error=response["error"],
                description=response.get("error_description", ""),
            )
            logline = lu.LOG_FMT.format(id=lu.get_session_id(context.state), message=msg)
            logger.debug(logline)
            raise SATOSAAuthenticationError(context.state, "Access denied")



import logging
from base64 import urlsafe_b64encode

from satosa.context import Context
from satosa.internal import InternalData

from .base import RequestMicroService
from ..exception import SATOSAConfigurationError
from ..exception import SATOSAError


logger = logging.getLogger(__name__)


class CustomRoutingError(SATOSAError):
    """SATOSA exception raised by CustomRouting rules"""
    pass


class DecideBackendByTargetIssuer(RequestMicroService):
    """
    Select target backend based on the target issuer.
    """

    def __init__(self, config:dict, *args, **kwargs):
        """
        Constructor.

        :param config: microservice configuration loaded from yaml file
        :type config: Dict[str, Dict[str, str]]
        """
        super().__init__(*args, **kwargs)

        self.target_mapping = config['target_mapping']
        self.default_backend = config['default_backend']

    def process(self, context:Context, data:InternalData):
        """Set context.target_backend based on the target issuer"""

        target_issuer = context.get_decoration(Context.KEY_TARGET_ENTITYID)
        if not target_issuer:
            logger.info('skipping backend decision because no target_issuer was found')
            return super().process(context, data)

        target_backend = (
            self.target_mapping.get(target_issuer)
            or self.default_backend
        )

        report = {
            'msg': 'decided target backend by target issuer',
            'target_issuer': target_issuer,
            'target_backend': target_backend,
        }
        logger.info(report)

        context.target_backend = target_backend
        return super().process(context, data)


class DecideBackendByRequester(RequestMicroService):
    """
    Select which backend should be used based on who the requester is.
    """

    def __init__(self, config, *args, **kwargs):
        """
        Constructor.
        :param config: mapping from requester identifier to
        backend module name under the key 'requester_mapping'.
        May also include default backend under key 'default_backend'.
        :type config: Dict[str, Dict[str, str]]
        """
        super().__init__(*args, **kwargs)
        self.requester_mapping = config['requester_mapping']
        self.default_backend = config.get('default_backend')

    def process(self, context, data):
        """
        Will modify the context.target_backend attribute based on the requester identifier.
        :param context: request context
        :param data: the internal request
        """
        context.target_backend = self.requester_mapping.get(data.requester) or self.default_backend
        return super().process(context, data)


class DecideIfRequesterIsAllowed(RequestMicroService):
    """
    Decide whether a requester is allowed to send an authentication request to the target entity.

    This micro service currently only works when a target entityid is set.
    Currently, a target entityid is set only when the `SAMLMirrorFrontend` is
    used.
    """
    def __init__(self, config, *args, **kwargs):
        super().__init__(*args, **kwargs)

        for target_entity, rules in config["rules"].items():
            conflicting_rules = set(rules.get("deny", [])).intersection(rules.get("allow", []))
            if conflicting_rules:
                raise SATOSAConfigurationError("Conflicting requester rules for DecideIfRequesterIsAllowed,"
                                               "{} is both denied and allowed".format(conflicting_rules))

        # target entity id is base64 url encoded to make it usable in URLs,
        # so we convert the rules the use those encoded entity id's instead
        self.rules = {self._b64_url(k): v for k, v in config["rules"].items()}

    def _b64_url(self, data):
        return urlsafe_b64encode(data.encode("utf-8")).decode("utf-8")

    def process(self, context, data):
        target_entity_id = context.get_decoration(Context.KEY_TARGET_ENTITYID)
        if None is target_entity_id:
            msg = "{name} can only be used when a target entityid is set".format(
                name=self.__class__.__name__
            )
            logger.error(msg)
            raise SATOSAError(msg)

        target_specific_rules = self.rules.get(target_entity_id)
        # default to allowing everything if there are no specific rules
        if not target_specific_rules:
            logger.debug("Requester '{}' allowed by default to target entity '{}' due to no entity specific rules".format(
                data.requester, target_entity_id
            ))
            return super().process(context, data)

        # deny rules takes precedence
        deny_rules = target_specific_rules.get("deny", [])
        if data.requester in deny_rules:
            logger.debug("Requester '{}' is not allowed by target entity '{}' due to deny rules '{}'".format(
                data.requester, target_entity_id, deny_rules
            ))
            raise SATOSAError("Requester is not allowed by target provider")

        allow_rules = target_specific_rules.get("allow", [])
        allow_all = "*" in allow_rules
        if data.requester in allow_rules or allow_all:
            logger.debug("Requester '{}' allowed by target entity '{}' due to allow rules '{}".format(
                data.requester, target_entity_id, allow_rules
            ))
            return super().process(context, data)

        logger.debug("Requester '{}' is not allowed by target entity '{}' due to final deny all rule in '{}'".format(
            data.requester, target_entity_id, deny_rules
        ))
        raise SATOSAError("Requester is not allowed by target provider")



import re
from chevron import render as render_mustache

from .base import ResponseMicroService
from ..util import get_dict_defaults


class MustachAttrValue(object):
    def __init__(self, attr_name, values):
        self._attr_name = attr_name
        self._values = values
        if any(['@' in v for v in values]):
            local_parts = []
            domain_parts = []
            scopes = dict()
            for v in values:
                (local_part, sep, domain_part) = v.partition('@')
                # probably not needed now...
                local_parts.append(local_part)
                domain_parts.append(domain_part)
                scopes[domain_part] = True
            self._scopes = list(scopes.keys())
        else:
            self._scopes = None

    def __str__(self):
        return ";".join(self._values)

    @property
    def values(self):
        return [{self._attr_name: v} for v in self._values]

    @property
    def value(self):
        if len(self._values) == 1:
            return self._values[0]
        else:
            return self._values

    @property
    def first(self):
        if len(self._values) > 0:
            return self._values[0]
        else:
            return ""

    @property
    def scope(self):
        if self._scopes is not None:
            return self._scopes[0]
        return ""


class AddSyntheticAttributes(ResponseMicroService):
    """
A class that add generated or synthetic attributes to a response set. Attribute
generation is done using mustach (http://mustache.github.io) templates. The
following example configuration illustrates most common features:

```yaml
module: satosa.micro_services.attribute_generation.AddSyntheticAttributes
name: AddSyntheticAttributes
config:
    synthetic_attributes:
        requester1:
            target_provider1:
                eduPersonAffiliation: member;employee
        default:
            default:
                schacHomeOrganization: {{eduPersonPrincipalName.scope}}
                schacHomeOrganizationType: tomfoolery provider

```

The use of "" and 'default' is synonymous. Attribute rules are not
overloaded or inherited. For instance a response for "requester1"
from target_provider1 in the above config will generate a (static) attribute
set of 'member' and 'employee' for the eduPersonAffiliation attribute
and nothing else. Note that synthetic attributes override existing
attributes if present.

*Evaluating and interpreting templates*

Attribute values are split on combinations of ';' and newline so that
a template resulting in the following text:
```
a;
b;c
```
results in three attribute values: 'a','b' and 'c'. Templates are
evaluated with a single context that represents the response attributes
before the microservice is processed. De-referencing the attribute
name as in '{{name}}' results in a ';'-separated list of all attribute
values. This notation is useful when you know there is only a single
attribute value in the set.

*Special contexts*

For treating the values as a list - eg for interating using mustach,
use the .values sub-context For instance to synthesize all first-last
name combinations do this:

```
{{#givenName.values}}
   {{#sn.values}}{{givenName}} {{sn}}{{/sn.values}}
{{/givenName.values}}
```

Note that the .values sub-context behaves as if it is an iterator
over single-value context with the same key name as the original
attribute name.

The .scope sub-context evalues to the right-hand part of any @
sign. This is assumed to be single valued.

The .first sub-context evalues to the first value of a context
which may be safer to use if the attribute is multivalued but
you don't care which value is used in a template.
    """

    def __init__(self, config, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.synthetic_attributes = config["synthetic_attributes"]

    def _synthesize(self, attributes, requester, provider):
        syn_attributes = dict()
        context = dict()

        for attr_name, values in attributes.items():
            context[attr_name] = MustachAttrValue(
                attr_name,
                values
                if values
                and isinstance(values, list)
                and all(isinstance(value, str) for value in values)
                else [],
            )

        recipes = get_dict_defaults(self.synthetic_attributes, requester, provider)
        for attr_name, fmt in recipes.items():
            syn_attributes[attr_name] = [
                value
                for token in re.split("[;\n]+", render_mustache(fmt, context))
                for value in [token.strip().strip(';')]
                if value
            ]
        return syn_attributes

    def process(self, context, data):
        data.attributes.update(self._synthesize(data.attributes, data.requester, data.auth_info.issuer))
        return super().process(context, data)



"""
This module contains methods to load, verify and build configurations for the satosa proxy.
"""
import logging
import os
import os.path

from satosa.exception import SATOSAConfigurationError
from satosa.yaml import load as yaml_load
from satosa.yaml import YAMLError


logger = logging.getLogger(__name__)


class SATOSAConfig(object):
    """
    A configuration class for the satosa proxy. Verifies that the given config holds all the
    necessary parameters.
    """
    sensitive_dict_keys = ["STATE_ENCRYPTION_KEY"]
    mandatory_dict_keys = ["BASE", "BACKEND_MODULES", "FRONTEND_MODULES",
                           "INTERNAL_ATTRIBUTES", "COOKIE_STATE_NAME"]

    def __init__(self, config):
        """
        Reads a given config and builds the SATOSAConfig.

        :type config: str | dict
        :rtype: satosa.satosa_config.SATOSAConfig

        :param config: Can be a file path or a dictionary
        :return: A verified SATOSAConfig
        """
        parsers = [self._load_dict, self._load_yaml]
        for parser in parsers:
            self._config = parser(config)
            if self._config is not None:
                break

        # Load sensitive config from environment variables
        for key in SATOSAConfig.sensitive_dict_keys:
            val = os.environ.get(f"SATOSA_{key}")
            if val:
                self._config[key] = val

        self._verify_dict(self._config)

        # Read plugin configs from dict or file path
        for key in ["BACKEND_MODULES", "FRONTEND_MODULES", "MICRO_SERVICES"]:
            plugin_configs = []
            for config in self._config.get(key, []):
                for parser in parsers:
                    plugin_config = parser(config)
                    if plugin_config:
                        plugin_configs.append(plugin_config)
                        break
                else:
                    raise SATOSAConfigurationError(f"Failed to load plugin config '{config}'")
            self._config[key] = plugin_configs

        for parser in parsers:
            _internal_attributes = parser(self._config["INTERNAL_ATTRIBUTES"])
            if _internal_attributes is not None:
                self._config["INTERNAL_ATTRIBUTES"] = _internal_attributes
                break
        if not self._config["INTERNAL_ATTRIBUTES"]:
            raise SATOSAConfigurationError("Could not load attribute mapping from 'INTERNAL_ATTRIBUTES.")

    def _verify_dict(self, conf):
        """
        Check that the configuration contains all necessary keys.

        :type conf: dict
        :rtype: None
        :raise SATOSAConfigurationError: if the configuration is incorrect

        :param conf: config to verify
        :return: None
        """
        if not conf:
            raise SATOSAConfigurationError("Missing configuration or unknown format")

        for key in SATOSAConfig.mandatory_dict_keys:
            if key not in conf:
                raise SATOSAConfigurationError("Missing key '%s' in config" % key)

        for key in SATOSAConfig.sensitive_dict_keys:
            if key not in conf and f"SATOSA_{key}" not in os.environ:
                raise SATOSAConfigurationError(f"Missing key '{key}' from config and ENVIRONMENT")

    def __getitem__(self, item):
        """
        Returns data bound to the key 'item'.

        :type item: str
        :rtype object

        :param item: key to data
        :return: data bound to key 'item'
        """
        return self._config[item]

    def __setitem__(self, key, value):
        """
        Inserts value into internal dict

        :type key: str
        :type value: object

        :param key: key
        :param value: data
        :return: None
        """
        self._config[key] = value

    def __contains__(self, key):
        return key in self._config

    def get(self, item, default=None):
        return self._config.get(item, default)

    def _load_dict(self, config):
        """
        Load config from dict

        :type config: dict
        :rtype: dict

        :param config: config to load
        :return: Loaded config
        """
        if isinstance(config, dict):
            return config

        return None

    def _load_yaml(self, config_file):
        """
        Load config from yaml file or string

        :type config_file: str
        :rtype: dict

        :param config_file: config to load. Can be file path or yaml string
        :return: Loaded config
        """

        try:
            with open(os.path.abspath(config_file)) as f:
                return yaml_load(f.read())
        except YAMLError as exc:
            logger.error("Could not parse config as YAML: {}".format(exc))
            if hasattr(exc, 'problem_mark'):
                mark = exc.problem_mark
                logger.error("Error position: ({line}:{column})".format(line=mark.line + 1, column=mark.column + 1))
        except IOError as e:
            logger.error("Could not open config file: {}".format(e))

        return None



"""
A saml2 backend module for the satosa proxy
"""
import copy
import functools
import json
import logging
import warnings as _warnings
from base64 import urlsafe_b64encode
from urllib.parse import urlparse

from saml2 import BINDING_HTTP_REDIRECT
from saml2.client import Saml2Client
from saml2.config import SPConfig
from saml2.extension.mdui import NAMESPACE as UI_NAMESPACE
from saml2.metadata import create_metadata_string
from saml2.authn_context import requested_authn_context
from saml2.samlp import RequesterID
from saml2.samlp import Scoping

import satosa.logging_util as lu
import satosa.util as util
from satosa.base import SAMLBaseModule
from satosa.base import SAMLEIDASBaseModule
from satosa.base import STATE_KEY as STATE_KEY_BASE
from satosa.context import Context
from satosa.internal import AuthenticationInformation
from satosa.internal import InternalData
from satosa.exception import SATOSAAuthenticationError
from satosa.exception import SATOSAMissingStateError
from satosa.exception import SATOSAAuthenticationFlowError
from satosa.response import SeeOther, Response
from satosa.saml_util import make_saml_response
from satosa.metadata_creation.description import (
    MetadataDescription, OrganizationDesc, ContactPersonDesc, UIInfoDesc
)
from satosa.backends.base import BackendModule


logger = logging.getLogger(__name__)


def get_memorized_idp(context, config, force_authn):
    memorized_idp = (
        config.get(SAMLBackend.KEY_MEMORIZE_IDP)
        and context.state.get(Context.KEY_MEMORIZED_IDP)
    )
    use_when_force_authn = config.get(
        SAMLBackend.KEY_USE_MEMORIZED_IDP_WHEN_FORCE_AUTHN
    )
    value = (not force_authn or use_when_force_authn) and memorized_idp
    return value


def get_force_authn(context, config, sp_config):
    """
    Return the force_authn value.

    The value comes from one of three place:
    - the configuration of the backend
    - the context, as it came through in the AuthnRequest handled by the frontend.
      note: the frontend should have been set to mirror the force_authn value.
    - the cookie, as it has been stored by the proxy on a redirect to the DS
      note: the frontend should have been set to mirror the force_authn value.

    The value is either "true" or None
    """
    mirror = config.get(SAMLBackend.KEY_MIRROR_FORCE_AUTHN)
    from_state = mirror and context.state.get(Context.KEY_FORCE_AUTHN)
    from_context = (
        mirror and context.get_decoration(Context.KEY_FORCE_AUTHN) in ["true", "1"]
    )
    from_config = sp_config.getattr("force_authn", "sp")
    is_set = str(from_state or from_context or from_config).lower() == "true"
    value = "true" if is_set else None
    return value


class SAMLBackend(BackendModule, SAMLBaseModule):
    """
    A saml2 backend module (acting as a SP).
    """
    KEY_DISCO_SRV = 'disco_srv'
    KEY_SAML_DISCOVERY_SERVICE_URL = 'saml_discovery_service_url'
    KEY_SAML_DISCOVERY_SERVICE_POLICY = 'saml_discovery_service_policy'
    KEY_SP_CONFIG = 'sp_config'
    KEY_SEND_REQUESTER_ID = 'send_requester_id'
    KEY_MIRROR_FORCE_AUTHN = 'mirror_force_authn'
    KEY_IS_PASSIVE = 'is_passive'
    KEY_MEMORIZE_IDP = 'memorize_idp'
    KEY_USE_MEMORIZED_IDP_WHEN_FORCE_AUTHN = 'use_memorized_idp_when_force_authn'

    VALUE_ACR_COMPARISON_DEFAULT = 'exact'

    def __init__(self, outgoing, internal_attributes, config, base_url, name):
        """
        :type outgoing:
        (satosa.context.Context, satosa.internal.InternalData) -> satosa.response.Response
        :type internal_attributes: dict[str, dict[str, list[str] | str]]
        :type config: dict[str, Any]
        :type base_url: str
        :type name: str

        :param outgoing: Callback should be called by the module after
                                   the authorization in the backend is done.
        :param internal_attributes: Internal attribute map
        :param config: The module config
        :param base_url: base url of the service
        :param name: name of the plugin
        """
        super().__init__(outgoing, internal_attributes, base_url, name)
        self.config = self.init_config(config)

        self.discosrv = config.get(SAMLBackend.KEY_DISCO_SRV)
        self.encryption_keys = []
        self.outstanding_queries = {}
        self.idp_blacklist_file = config.get('idp_blacklist_file', None)

        sp_config = SPConfig().load(copy.deepcopy(config[SAMLBackend.KEY_SP_CONFIG]))

        # if encryption_keypairs is defined, use those keys for decryption
        # else, if key_file and cert_file are defined, use them for decryption
        # otherwise, do not use any decryption key.
        # ensure the choice is reflected back in the configuration.
        sp_conf_encryption_keypairs = sp_config.getattr('encryption_keypairs', '')
        sp_conf_key_file = sp_config.getattr('key_file', '')
        sp_conf_cert_file = sp_config.getattr('cert_file', '')
        sp_keypairs = (
            sp_conf_encryption_keypairs
            if sp_conf_encryption_keypairs
            else [{'key_file': sp_conf_key_file, 'cert_file': sp_conf_cert_file}]
            if sp_conf_key_file and sp_conf_cert_file
            else []
        )
        sp_config.setattr('', 'encryption_keypairs', sp_keypairs)

        # load the encryption keys
        key_file_paths = [pair['key_file'] for pair in sp_keypairs]
        for p in key_file_paths:
            with open(p) as key_file:
                self.encryption_keys.append(key_file.read())

        # finally, initialize the client object
        self.sp = Saml2Client(sp_config)

    def get_idp_entity_id(self, context):
        """
        :type context: satosa.context.Context
        :rtype: str | None

        :param context: The current context
        :return: the entity_id of the idp or None
        """

        idps = self.sp.metadata.identity_providers()
        only_one_idp_in_metadata = (
            "mdq" not in self.config["sp_config"]["metadata"]
            and len(idps) == 1
        )

        only_idp = only_one_idp_in_metadata and idps[0]
        target_entity_id = context.get_decoration(Context.KEY_TARGET_ENTITYID)
        force_authn = get_force_authn(context, self.config, self.sp.config)
        memorized_idp = get_memorized_idp(context, self.config, force_authn)
        entity_id = only_idp or target_entity_id or memorized_idp or None

        msg = {
            "message": "Selected IdP",
            "only_one": only_idp,
            "target_entity_id": target_entity_id,
            "force_authn": force_authn,
            "memorized_idp": memorized_idp,
            "entity_id": entity_id,
        }
        logline = lu.LOG_FMT.format(id=lu.get_session_id(context.state), message=msg)
        logger.info(logline)
        return entity_id

    def start_auth(self, context, internal_req):
        """
        See super class method satosa.backends.base.BackendModule#start_auth

        :type context: satosa.context.Context
        :type internal_req: satosa.internal.InternalData
        :rtype: satosa.response.Response
        """

        entity_id = self.get_idp_entity_id(context)
        if entity_id is None:
            # since context is not passed to disco_query
            # keep the information in the state cookie
            context.state[Context.KEY_FORCE_AUTHN] = get_force_authn(
                context, self.config, self.sp.config
            )
            return self.disco_query(context)

        return self.authn_request(context, entity_id)

    def disco_query(self, context):
        """
        Makes a request to the discovery server

        :type context: satosa.context.Context
        :type internal_req: satosa.internal.InternalData
        :rtype: satosa.response.SeeOther

        :param context: The current context
        :param internal_req: The request
        :return: Response
        """
        endpoints = self.sp.config.getattr("endpoints", "sp")
        return_url = endpoints["discovery_response"][0][0]

        disco_url = (
            context.get_decoration(SAMLBackend.KEY_SAML_DISCOVERY_SERVICE_URL)
            or self.discosrv
        )
        disco_policy = context.get_decoration(
            SAMLBackend.KEY_SAML_DISCOVERY_SERVICE_POLICY
        )

        args = {"return": return_url}
        if disco_policy:
            args["policy"] = disco_policy

        loc = self.sp.create_discovery_service_request(
            disco_url, self.sp.config.entityid, **args
        )

        msg = {
            "message": "Sending user to the discovery service",
            "disco_url": loc
        }
        logline = lu.LOG_FMT.format(id=lu.get_session_id(context.state), message=msg)
        logger.info(logline)

        return SeeOther(loc)

    def construct_requested_authn_context(self, entity_id, *, target_accr=None):
        acr_entry = (
            target_accr
            or util.get_dict_defaults(self.acr_mapping or {}, entity_id)
        )
        if not acr_entry:
            return None

        if type(acr_entry) is not dict:
            acr_entry = {
                "class_ref": acr_entry,
                "comparison": self.VALUE_ACR_COMPARISON_DEFAULT,
            }

        authn_context = requested_authn_context(
            acr_entry['class_ref'], comparison=acr_entry.get(
                'comparison', self.VALUE_ACR_COMPARISON_DEFAULT
            )
        )

        return authn_context

    def authn_request(self, context, entity_id):
        """
        Do an authorization request on idp with given entity id.
        This is the start of the authorization.

        :type context: satosa.context.Context
        :type entity_id: str
        :rtype: satosa.response.Response

        :param context: The current context
        :param entity_id: Target IDP entity id
        :return: response to the user agent
        """

        # If IDP blacklisting is enabled and the selected IDP is blacklisted,
        # stop here
        if self.idp_blacklist_file:
            with open(self.idp_blacklist_file) as blacklist_file:
                blacklist_array = json.load(blacklist_file)['blacklist']
                if entity_id in blacklist_array:
                    msg = {
                        "message": "AuthnRequest Failed",
                        "error": f"Selected IdP with EntityID {entity_id} is blacklisted for this backend",
                    }
                    logline = lu.LOG_FMT.format(id=lu.get_session_id(context.state), message=msg)
                    logger.info(logline)
                    raise SATOSAAuthenticationError(context.state, msg)

        kwargs = {}
        target_accr = context.state.get(Context.KEY_TARGET_AUTHN_CONTEXT_CLASS_REF)
        authn_context = self.construct_requested_authn_context(entity_id, target_accr=target_accr)
        if authn_context:
            kwargs["requested_authn_context"] = authn_context
        if self.config.get(SAMLBackend.KEY_MIRROR_FORCE_AUTHN):
            kwargs["force_authn"] = get_force_authn(
                context, self.config, self.sp.config
            )
        if self.config.get(SAMLBackend.KEY_SEND_REQUESTER_ID):
            requester = context.state.state_dict[STATE_KEY_BASE]['requester']
            kwargs["scoping"] = Scoping(requester_id=[RequesterID(text=requester)])
        if self.config.get(SAMLBackend.KEY_IS_PASSIVE):
            kwargs["is_passive"] = "true"

        try:
            acs_endp, response_binding = self._get_acs(context)
            relay_state = util.rndstr()
            req_id, binding, http_info = self.sp.prepare_for_negotiated_authenticate(
                entityid=entity_id,
                assertion_consumer_service_url=acs_endp,
                response_binding=response_binding,
                relay_state=relay_state,
                **kwargs,
            )
        except Exception as e:
            msg = {
                "message": "AuthnRequest Failed",
                "error": f"Failed to construct the AuthnRequest for state: {e}",
            }
            logline = lu.LOG_FMT.format(id=lu.get_session_id(context.state), message=msg)
            logger.info(logline)
            raise SATOSAAuthenticationError(context.state, msg) from e

        if self.sp.config.getattr('allow_unsolicited', 'sp') is False:
            if req_id in self.outstanding_queries:
                msg = {
                    "message": "AuthnRequest Failed",
                    "error": f"Request with duplicate id {req_id}",
                }
                logline = lu.LOG_FMT.format(id=lu.get_session_id(context.state), message=msg)
                logger.info(logline)
                raise SATOSAAuthenticationError(context.state, msg)
            self.outstanding_queries[req_id] = req_id

        context.state[self.name] = {"relay_state": relay_state}
        return make_saml_response(binding, http_info)

    def _get_acs(self, context):
        """
        Select the AssertionConsumerServiceURL and binding.

        :param context: The current context
        :type context: satosa.context.Context
        :return: Selected ACS URL and binding
        :rtype: tuple(str, str)
        """
        acs_strategy = self.config.get("acs_selection_strategy", "use_first_acs")
        if acs_strategy == "use_first_acs":
            acs_strategy_fn = self._use_first_acs
        elif acs_strategy == "prefer_matching_host":
            acs_strategy_fn = self._prefer_matching_host
        else:
            msg = "Invalid value for '{}' ({}). Using the first ACS instead".format(
                "acs_selection_strategy", acs_strategy
            )
            logger.error(msg)
            acs_strategy_fn = self._use_first_acs
        return acs_strategy_fn(context)

    def _use_first_acs(self, context):
        return self.sp.config.getattr("endpoints", "sp")["assertion_consumer_service"][
            0
        ]

    def _prefer_matching_host(self, context):
        acs_config = self.sp.config.getattr("endpoints", "sp")[
            "assertion_consumer_service"
        ]
        try:
            hostname = context.http_headers["HTTP_HOST"]
            for acs, binding in acs_config:
                parsed_acs = urlparse(acs)
                if hostname == parsed_acs.netloc:
                    msg = "Selected ACS '{}' based on the request".format(acs)
                    logline = lu.LOG_FMT.format(
                        id=lu.get_session_id(context.state), message=msg
                    )
                    logger.debug(logline)
                    return acs, binding
        except (TypeError, KeyError):
            pass

        msg = "Can't find an ACS URL to this hostname ({}), selecting the first one".format(
            context.http_headers.get("HTTP_HOST", "") if context.http_headers else ""
        )
        logline = lu.LOG_FMT.format(id=lu.get_session_id(context.state), message=msg)
        logger.debug(logline)
        return self._use_first_acs(context)

    def authn_response(self, context, binding):
        """
        Endpoint for the idp response
        :type context: satosa.context,Context
        :type binding: str
        :rtype: satosa.response.Response

        :param context: The current context
        :param binding: The saml binding type
        :return: response
        """

        if self.name not in context.state:
            """
            If we end up here, it means that the user returns to the proxy
            without the SATOSA session cookie. This can happen at least in the
            following cases:
            - the user deleted the cookie from the browser
            - the browser of the user blocked the cookie
            - the user has completed an authentication flow, the cookie has
              been removed by SATOSA and then the user used the back button
              of their browser and resend the authentication response, but
              without the SATOSA session cookie
            """
            msg = {
                "message": "Authentication failed",
                "error": "Received AuthN response without a SATOSA session cookie",
            }
            logline = lu.LOG_FMT.format(id=lu.get_session_id(context.state), message=msg)
            logger.info(logline)
            raise SATOSAMissingStateError(msg)

        samlresponse = context.request.get("SAMLResponse")
        if not samlresponse:
            msg = {
                "message": "Authentication failed",
                "error": "SAML Response not found in context.request",
            }
            logline = lu.LOG_FMT.format(id=lu.get_session_id(context.state), message=msg)
            logger.info(logline)
            raise SATOSAAuthenticationError(context.state, msg)

        try:
            authn_response = self.sp.parse_authn_request_response(
                samlresponse, binding, outstanding=self.outstanding_queries
            )
        except Exception as e:
            msg = {
                "message": "Authentication failed",
                "error": f"Failed to parse Authn response: {e}",
            }
            logline = lu.LOG_FMT.format(id=lu.get_session_id(context.state), message=msg)
            logger.debug(logline, exc_info=True)
            logger.info(logline)
            raise SATOSAAuthenticationError(context.state, msg) from e

        if self.sp.config.getattr('allow_unsolicited', 'sp') is False:
            req_id = authn_response.in_response_to
            if req_id not in self.outstanding_queries:
                msg = {
                    "message": "Authentication failed",
                    "error": f"No corresponding request with id: {req_id}",
                }
                logline = lu.LOG_FMT.format(id=lu.get_session_id(context.state), message=msg)
                logger.info(logline)
                raise SATOSAAuthenticationError(context.state, msg)
            del self.outstanding_queries[req_id]

        # check if the relay_state matches the cookie state
        if context.state[self.name].get("relay_state") != context.request["RelayState"]:
            msg = {
                "message": "Authentication failed",
                "error": "Response state query param did not match relay state for request",
            }
            logline = lu.LOG_FMT.format(id=lu.get_session_id(context.state), message=msg)
            logger.info(logline)
            raise SATOSAAuthenticationError(context.state, msg)

        context.decorate(Context.KEY_METADATA_STORE, self.sp.metadata)
        if self.config.get(SAMLBackend.KEY_MEMORIZE_IDP):
            issuer = authn_response.response.issuer.text.strip()
            context.state[Context.KEY_MEMORIZED_IDP] = issuer
        context.state.pop(Context.KEY_FORCE_AUTHN, None)
        return self.auth_callback_func(context, self._translate_response(authn_response, context.state))

    def disco_response(self, context):
        """
        Endpoint for the discovery server response

        :type context: satosa.context.Context
        :rtype: satosa.response.Response

        :param context: The current context
        :return: response
        """
        info = context.request
        state = context.state

        if 'SATOSA_BASE' not in state:
            raise SATOSAAuthenticationFlowError("Discovery response without AuthN request")

        entity_id = info.get("entityID")
        msg = {
            "message": "Received response from the discovery service",
            "entity_id": entity_id,
        }
        logline = lu.LOG_FMT.format(id=lu.get_session_id(context.state), message=msg)
        logger.info(logline)
        if not entity_id:
            raise SATOSAAuthenticationError(state, msg) from err

        return self.authn_request(context, entity_id)

    def _translate_response(self, response, state):
        """
        Translates a saml authorization response to an internal response

        :type response: saml2.response.AuthnResponse
        :rtype: satosa.internal.InternalData
        :param response: The saml authorization response
        :return: A translated internal response
        """

        # The response may have been encrypted by the IdP so if we have an
        # encryption key, try it.
        if self.encryption_keys:
            response.parse_assertion(keys=self.encryption_keys)

        issuer = response.response.issuer.text
        authn_context_ref, authenticating_authorities, authn_instant = next(
            iter(response.authn_info()), [None, None, None]
        )
        authenticating_authority = (
            authenticating_authorities[-1]
            if authenticating_authorities
            else None
        )
        auth_info = AuthenticationInformation(
            auth_class_ref=authn_context_ref,
            timestamp=authn_instant,
            authority=authenticating_authority,
            issuer=issuer,
        )

        # The SAML response may not include a NameID.
        subject = response.get_subject()
        name_id = subject.text if subject else None
        name_id_format = subject.format if subject else None

        attributes = self.converter.to_internal(
            self.attribute_profile, response.ava,
        )

        internal_resp = InternalData(
            auth_info=auth_info,
            attributes=attributes,
            subject_type=name_id_format,
            subject_id=name_id,
        )

        msg = "backend received attributes: {}".format(response.ava)
        logline = lu.LOG_FMT.format(id=lu.get_session_id(state), message=msg)
        logger.debug(logline)

        msg = {
            "message": "Attributes received by the backend",
            "issuer": issuer,
            "attributes": " ".join(list(response.ava.keys()))
        }
        if name_id_format:
            msg['name_id'] = name_id_format
        logline = lu.LOG_FMT.format(id=lu.get_session_id(state), message=msg)
        logger.info(logline)

        return internal_resp

    def _metadata_endpoint(self, context):
        """
        Endpoint for retrieving the backend metadata
        :type context: satosa.context.Context
        :rtype: satosa.response.Response

        :param context: The current context
        :return: response with metadata
        """
        msg = "Sending metadata response for entityId = {}".format(self.sp.config.entityid)
        logline = lu.LOG_FMT.format(id=lu.get_session_id(context.state), message=msg)
        logger.debug(logline)

        metadata_string = create_metadata_string(
            configfile=None, config=self.sp.config, valid=4
        ).decode("utf-8")
        return Response(metadata_string, content="text/xml")

    def register_endpoints(self):
        """
        See super class method satosa.backends.base.BackendModule#register_endpoints
        :rtype list[(str, ((satosa.context.Context, Any) -> Any, Any))]
        """
        url_map = []
        sp_endpoints = self.sp.config.getattr("endpoints", "sp")
        for endp, binding in sp_endpoints["assertion_consumer_service"]:
            parsed_endp = urlparse(endp)
            url_map.append(("^%s$" % parsed_endp.path[1:], functools.partial(self.authn_response, binding=binding)))
            if binding == BINDING_HTTP_REDIRECT:
                msg = " ".join(
                    [
                        "AssertionConsumerService endpoint with binding",
                        BINDING_HTTP_REDIRECT,
                        "is not recommended.",
                        "Quoting section 4.1.2 of",
                        "'Profiles for the OASIS Security Assertion Markup Language (SAML) V2.0':",
                        "The HTTP Redirect binding MUST NOT be used,",
                        "as the response will typically exceed the URL length",
                        "permitted by most user agents.",
                    ]
                )
                _warnings.warn(msg, UserWarning)

        if self.discosrv:
            for endp, binding in sp_endpoints["discovery_response"]:
                parsed_endp = urlparse(endp)
                url_map.append(
                    ("^%s$" % parsed_endp.path[1:], self.disco_response))

        if self.expose_entityid_endpoint():
            logger.debug("Exposing backend entity endpoint = {}".format(self.sp.config.entityid))
            parsed_entity_id = urlparse(self.sp.config.entityid)
            url_map.append(("^{0}".format(parsed_entity_id.path[1:]),
                            self._metadata_endpoint))

        if self.enable_metadata_reload():
            url_map.append(
                ("^%s/%s$" % (self.name, "reload-metadata"), self._reload_metadata))

        return url_map

    def _reload_metadata(self, context):
        """
        Reload SAML metadata
        """
        logger.debug("Reloading metadata")
        res = self.sp.reload_metadata(
            copy.deepcopy(self.config[SAMLBackend.KEY_SP_CONFIG]['metadata'])
        )
        message = "Metadata reload %s" % ("OK" if res else "failed")
        status = "200 OK" if res else "500 FAILED"
        return Response(message=message, status=status)

    def get_metadata_desc(self):
        """
        See super class satosa.backends.backend_base.BackendModule#get_metadata_desc
        :rtype: satosa.metadata_creation.description.MetadataDescription
        """
        entity_descriptions = []

        idp_entities = self.sp.metadata.with_descriptor("idpsso")
        for entity_id, entity in idp_entities.items():
            description = MetadataDescription(urlsafe_b64encode(entity_id.encode("utf-8")).decode("utf-8"))

            # Add organization info
            try:
                organization_info = entity["organization"]
            except KeyError:
                pass
            else:
                organization = OrganizationDesc()
                for name_info in organization_info.get("organization_name", []):
                    organization.add_name(name_info["text"], name_info["lang"])
                for display_name_info in organization_info.get("organization_display_name", []):
                    organization.add_display_name(display_name_info["text"], display_name_info["lang"])
                for url_info in organization_info.get("organization_url", []):
                    organization.add_url(url_info["text"], url_info["lang"])
                description.organization = organization

            # Add contact person info
            try:
                contact_persons = entity["contact_person"]
            except KeyError:
                pass
            else:
                for person in contact_persons:
                    person_desc = ContactPersonDesc()
                    person_desc.contact_type = person.get("contact_type")
                    for address in person.get('email_address', []):
                        person_desc.add_email_address(address["text"])
                    if "given_name" in person:
                        person_desc.given_name = person["given_name"]["text"]
                    if "sur_name" in person:
                        person_desc.sur_name = person["sur_name"]["text"]

                    description.add_contact_person(person_desc)

            # Add UI info
            ui_info = self.sp.metadata.extension(entity_id, "idpsso_descriptor", "{}&UIInfo".format(UI_NAMESPACE))
            if ui_info:
                ui_info = ui_info[0]
                ui_info_desc = UIInfoDesc()
                for desc in ui_info.get("description", []):
                    ui_info_desc.add_description(desc["text"], desc["lang"])
                for name in ui_info.get("display_name", []):
                    ui_info_desc.add_display_name(name["text"], name["lang"])
                for logo in ui_info.get("logo", []):
                    ui_info_desc.add_logo(logo["text"], logo["width"], logo["height"], logo.get("lang"))
                for keywords in ui_info.get("keywords", []):
                    ui_info_desc.add_keywords(keywords.get("text", []), keywords.get("lang"))
                for information_url in ui_info.get("information_url", []):
                    ui_info_desc.add_information_url(information_url.get("text"), information_url.get("lang"))
                for privacy_statement_url in ui_info.get("privacy_statement_url", []):
                    ui_info_desc.add_privacy_statement_url(
                        privacy_statement_url.get("text"), privacy_statement_url.get("lang")
                    )
                description.ui_info = ui_info_desc

            entity_descriptions.append(description)
        return entity_descriptions


class SAMLEIDASBackend(SAMLBackend, SAMLEIDASBaseModule):
    """
    A saml2 eidas backend module (acting as a SP).
    """
    VALUE_ACR_CLASS_REF_DEFAULT = 'http://eidas.europa.eu/LoA/high'
    VALUE_ACR_COMPARISON_DEFAULT = 'minimum'

    def init_config(self, config):
        config = super().init_config(config)

        spec_eidas_sp = {
            'acr_mapping': {
                "": {
                    'class_ref': self.VALUE_ACR_CLASS_REF_DEFAULT,
                    'comparison': self.VALUE_ACR_COMPARISON_DEFAULT,
                },
            },
            'sp_config.service.sp.authn_requests_signed': True,
            'sp_config.service.sp.want_response_signed': True,
            'sp_config.service.sp.allow_unsolicited': False,
            'sp_config.service.sp.force_authn': True,
            'sp_config.service.sp.hide_assertion_consumer_service': True,
            'sp_config.service.sp.sp_type': ['private', 'public'],
            'sp_config.service.sp.sp_type_in_metadata': [True, False],
        }

        return util.check_set_dict_defaults(config, spec_eidas_sp)



from ..attribute_processor import AttributeProcessorError
from .base_processor import BaseProcessor

import hashlib


CONFIG_KEY_SALT = 'salt'
CONFIG_DEFAULT_SALT = ''
CONFIG_KEY_HASHALGO = 'hash_algo'
CONFIG_DEFAULT_HASHALGO = 'sha256'


class HashProcessor(BaseProcessor):
    def process(self, internal_data, attribute, **kwargs):
        salt = kwargs.get(CONFIG_KEY_HASHALGO, CONFIG_DEFAULT_SALT)
        hash_algo = kwargs.get(CONFIG_KEY_HASHALGO, CONFIG_DEFAULT_HASHALGO)
        if hash_algo not in hashlib.algorithms_available:
            raise AttributeProcessorError(
                "Hash algorithm not supported: {}".format(hash_algo))

        attributes = internal_data.attributes
        value = attributes.get(attribute, [None])[0]
        if value is None:
            raise AttributeProcessorError(
                "No value for attribute: {}".format(attribute))

        hasher = hashlib.new(hash_algo)
        hasher.update(value.encode('utf-8'))
        hasher.update(salt.encode('utf-8'))
        value_hashed = hasher.hexdigest()
        attributes[attribute][0] = value_hashed



import re

from .base import ResponseMicroService
from ..exception import SATOSAAuthenticationError
from ..util import get_dict_defaults

class AttributeAuthorization(ResponseMicroService):
    """
    A microservice that performs simple regexp-based authorization based on response
    attributes. There are two configuration options to match attribute values in order
    to allow or deny authorization.

    The configuration is wrapped in two nested dicts that specialize the options per
    requester (SP/RP) and issuer (IdP/OP).

    There are also two options to enforce presence of the attributes that are going to
    be checked.

    Example configuration:

      ```yaml
      config:
        force_attributes_presence_on_allow: true
        attribute_allow:
          target_provider1:
            requester1:
              attr1:
                - "^foo:bar$"
                - "^kaka$"
            default:
              attr1:
                - "plupp@.+$"
          "":
            "":
              attr2:
                - "^knytte:.*$"

        force_attributes_presence_on_deny: false
        attribute_deny:
          default:
            default:
              eppn:
                - "^[^@]+$"
      ```

    The use of "" and "default" is synonymous. Attribute rules are not overloaded
    or inherited. For instance a response from "provider2" would only be allowed
    through if the eppn attribute had all values containing an '@' (something
    perhaps best implemented via an allow rule in practice). Responses from
    target_provider1 bound for requester1 would be allowed through only if attr1
    contained foo:bar or kaka. Note that attribute filters (the leaves of the
    structure above) are ORed together - i.e any attribute match is sufficient.
    """

    def __init__(self, config, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.attribute_allow = config.get("attribute_allow", {})
        self.attribute_deny = config.get("attribute_deny", {})
        self.force_attributes_presence_on_allow = config.get("force_attributes_presence_on_allow", False)
        self.force_attributes_presence_on_deny = config.get("force_attributes_presence_on_deny", False)

    def _check_authz(self, context, attributes, requester, provider):
        for attribute_name, attribute_filters in get_dict_defaults(self.attribute_allow, requester, provider).items():
            attr_values = attributes.get(attribute_name)
            if attr_values is not None:
                if not any(
                    [
                        any(filter(lambda x: re.search(af, x), attr_values))
                        for af in attribute_filters
                    ]
                ):
                    raise SATOSAAuthenticationError(context.state, "Permission denied")
            elif self.force_attributes_presence_on_allow:
                raise SATOSAAuthenticationError(context.state, "Permission denied")

        for attribute_name, attribute_filters in get_dict_defaults(self.attribute_deny, requester, provider).items():
            attr_values = attributes.get(attribute_name)
            if attr_values is not None:
                if any(
                    [
                        any(filter(lambda x: re.search(af, x), attributes[attribute_name]))
                        for af in attribute_filters
                    ]
                ):
                    raise SATOSAAuthenticationError(context.state, "Permission denied")
            elif self.force_attributes_presence_on_deny:
                raise SATOSAAuthenticationError(context.state, "Permission denied")

    def process(self, context, data):
        self._check_authz(context, data.attributes, data.requester, data.auth_info.issuer)
        return super().process(context, data)



"""
An account linking module for the satosa proxy
"""
import json
import logging

import requests
from jwkest.jwk import rsa_load, RSAKey
from jwkest.jws import JWS

from satosa.internal import InternalData
from ..exception import SATOSAAuthenticationError
from ..micro_services.base import ResponseMicroService
from ..response import Redirect

import satosa.logging_util as lu
logger = logging.getLogger(__name__)


class AccountLinking(ResponseMicroService):
    """
    Module for handling account linking and recovery. Uses an external account linking service
    """

    def __init__(self, config, *args, **kwargs):
        """
        :type config: satosa.satosa_config.SATOSAConfig
        :param config: The SATOSA proxy config
        """
        super().__init__(*args, **kwargs)
        self.api_url = config["api_url"]
        self.redirect_url = config["redirect_url"]
        self.signing_key = RSAKey(key=rsa_load(config["sign_key"]), use="sig", alg="RS256")
        self.endpoint = "/handle_account_linking"
        self.id_to_attr = config.get("id_to_attr", None)
        logger.info("Account linking is active")

    def _handle_al_response(self, context):
        """
        Endpoint for handling account linking service response. When getting here
        user might have approved or rejected linking their account

        :type context: satosa.context.Context
        :rtype: satosa.response.Response

        :param context: The current context
        :return: response
        """
        saved_state = context.state[self.name]
        internal_response = InternalData.from_dict(saved_state)

        #subject_id here is the linked id , not the facebook one, Figure out what to do
        status_code, message = self._get_uuid(context, internal_response.auth_info.issuer, internal_response.attributes['issuer_user_id'])

        if status_code == 200:
            msg = "issuer/id pair is linked in AL service"
            logline = lu.LOG_FMT.format(id=lu.get_session_id(context.state), message=msg)
            logger.info(logline)
            internal_response.subject_id = message
            if self.id_to_attr:
                internal_response.attributes[self.id_to_attr] = [message]

            del context.state[self.name]
            return super().process(context, internal_response)
        else:
            # User selected not to link their accounts, so the internal.response.subject_id is based on the
            # issuers id/sub which is fine
            msg = "User selected to not link their identity in AL service"
            logline = lu.LOG_FMT.format(id=lu.get_session_id(context.state), message=msg)
            logger.info(logline)
            del context.state[self.name]
            return super().process(context, internal_response)


    def process(self, context, internal_response):
        """
        Manage account linking and recovery

        :type context: satosa.context.Context
        :type internal_response: satosa.internal.InternalData
        :rtype: satosa.response.Response

        :param context:
        :param internal_response:
        :return: response
        :
        """

        status_code, message = self._get_uuid(context, internal_response.auth_info.issuer, internal_response.subject_id)

        data = {
            "issuer": internal_response.auth_info.issuer,
            "redirect_endpoint": "%s/account_linking%s" % (self.base_url, self.endpoint)
        }

        # Store the issuer subject_id/sub because we'll need it in handle_al_response
        internal_response.attributes['issuer_user_id'] = internal_response.subject_id
        if status_code == 200:
            msg = "issuer/id pair is linked in AL service"
            logline = lu.LOG_FMT.format(id=lu.get_session_id(context.state), message=msg)
            logger.info(logline)
            internal_response.subject_id = message
            data['user_id'] = message
            if self.id_to_attr:
                internal_response.attributes[self.id_to_attr] = [message]
        else:
            msg = "issuer/id pair is not linked in AL service. Got a ticket"
            logline = lu.LOG_FMT.format(id=lu.get_session_id(context.state), message=msg)
            logger.info(logline)
            data['ticket'] = message
        jws = JWS(json.dumps(data), alg=self.signing_key.alg).sign_compact([self.signing_key])
        context.state[self.name] = internal_response.to_dict()
        return Redirect("%s/%s" % (self.redirect_url, jws))

    def _get_uuid(self, context, issuer, id):
        """
        Ask the account linking service for a uuid.
        If the given issuer/id pair is not linked, then the function will return a ticket.
        This ticket should be used for linking the issuer/id pair to the user account

        :type context: satosa.context.Context
        :type issuer: str
        :type id: str
        :rtype: (int, str)

        :param context: The current context
        :param issuer: the issuer used for authentication
        :param id: the given id
        :return: response status code and message
            (200, uuid) or (404, ticket)
        """
        data = {
            "idp": issuer,
            "id": id,
            "redirect_endpoint": "%s/account_linking%s" % (self.base_url, self.endpoint)
        }
        jws = JWS(json.dumps(data), alg=self.signing_key.alg).sign_compact([self.signing_key])

        try:
            request = "{}/get_id?jwt={}".format(self.api_url, jws)
            response = requests.get(request)
        except Exception as con_exc:
            msg = "Could not connect to account linking service"
            logline = lu.LOG_FMT.format(id=lu.get_session_id(context.state), message=msg)
            logger.critical(logline)
            raise SATOSAAuthenticationError(context.state, msg) from con_exc

        if response.status_code not in [200, 404]:
            msg = "Got status code '{}' from account linking service".format(response.status_code)
            logline = lu.LOG_FMT.format(id=lu.get_session_id(context.state), message=msg)
            logger.critical(logline)
            raise SATOSAAuthenticationError(context.state, msg)

        return response.status_code, response.text

    def register_endpoints(self):
        """
        Register consent module endpoints

        :rtype: list[(srt, (satosa.context.Context) -> Any)]

        :return: A list of endpoints bound to a function
        """
        return [("^account_linking%s$" % self.endpoint, self._handle_al_response)]



import json
import logging
import logging.config
from io import BytesIO
from urllib.parse import parse_qsl as _parse_query_string

from cookies_samesite_compat import CookiesSameSiteCompatMiddleware

import satosa

from .base import SATOSABase
from .context import Context
from .response import ServiceError
from .response import NotFound


logger = logging.getLogger(__name__)


def parse_query_string(data):
    query_param_pairs = _parse_query_string(data)
    query_param_dict = dict(query_param_pairs)
    if "resource" in query_param_dict:
        query_param_dict["resource"] = [t[1] for t in query_param_pairs if t[0] == "resource"]
    return query_param_dict


def unpack_get(environ):
    """
    Unpacks a redirect request query string.
    :param environ: whiskey application environment.
    :return: A dictionary with parameters.
    """
    return parse_query_string(environ.get("QUERY_STRING"))


def unpack_post(environ, content_length):
    """
    Unpacks a post request query string.
    :param environ: whiskey application environment.
    :return: A dictionary with parameters.
    """
    post_body = environ['wsgi.input'].read(content_length).decode("utf-8")
    data = None
    if "application/x-www-form-urlencoded" in environ["CONTENT_TYPE"]:
        data = parse_query_string(post_body)
    elif "application/json" in environ["CONTENT_TYPE"]:
        data = json.loads(post_body)

    logline = "unpack_post:: {}".format(data)
    logger.debug(logline)
    return data


def unpack_request(environ, content_length=0):
    """
    Unpacks a get or post request query string.
    :param environ: whiskey application environment.
    :return: A dictionary with parameters.
    """
    data = None
    if environ["REQUEST_METHOD"] == "GET":
        data = unpack_get(environ)
    elif environ["REQUEST_METHOD"] == "POST":
        data = unpack_post(environ, content_length)

    logline = "read request data: {}".format(data)
    logger.debug(logline)
    return data


def collect_server_headers(environ):
    headers = {
        header_name: header_value
        for header_name, header_value in environ.items()
        if header_name.startswith("SERVER_")
    }
    return headers


def collect_http_headers(environ):
    headers = {
        header_name: header_value
        for header_name, header_value in environ.items()
        if (
            header_name.startswith("HTTP_")
            or header_name.startswith("REMOTE_")
        )
    }
    return headers


class ToBytesMiddleware(object):
    """Converts a message to bytes to be sent by WSGI server."""

    def __init__(self, app):
        self.app = app

    def __call__(self, environ, start_response):
        data = self.app(environ, start_response)

        if isinstance(data, list):
            encoded_data = []
            for d in data:
                if isinstance(d, bytes):
                    encoded_data.append(d)
                else:
                    encoded_data.append(d.encode("utf-8"))
            return encoded_data

        if isinstance(data, str):
            return data.encode("utf-8")

        return data


class WsgiApplication(SATOSABase):
    def __init__(self, config):
        super().__init__(config)

    def __call__(self, environ, start_response, debug=False):
        path = environ.get('PATH_INFO', '').lstrip('/')
        if ".." in path or path == "":
            resp = NotFound("Couldn't find the page you asked for!")
            return resp(environ, start_response)

        context = Context()
        context.path = path

        # copy wsgi.input stream to allow it to be re-read later by satosa plugins
        # see: http://stackoverflow.com/questions/1783383/how-do-i-copy-wsgi-input-if-i-want-to-process-post-data-more-than-once
        content_length = int(environ.get('CONTENT_LENGTH', '0') or '0')
        body = BytesIO(environ['wsgi.input'].read(content_length))
        environ['wsgi.input'] = body

        context.request = unpack_request(environ, content_length)
        context.request_uri = environ.get("REQUEST_URI")
        context.request_method = environ.get("REQUEST_METHOD")
        context.qs_params = parse_query_string(environ.get("QUERY_STRING"))
        context.server = collect_server_headers(environ)
        context.http_headers = collect_http_headers(environ)
        context.cookie = context.http_headers.get("HTTP_COOKIE", "")
        context.request_authorization = context.http_headers.get("HTTP_AUTHORIZATION", "")

        environ['wsgi.input'].seek(0)

        logline = {
            "message": "Proxy server received request",
            "request_method": context.request_method,
            "request_uri": context.request_uri,
            "content_length": content_length,
            "request_data": context.request,
            "query_params": context.qs_params,
            "http_headers": context.http_headers,
            "server_headers": context.server,
        }
        logger.debug(logline)

        try:
            resp = self.run(context)
            if isinstance(resp, Exception):
                raise resp
            return resp(environ, start_response)
        except Exception as e:
            logger.exception(str(e))
            if debug:
                raise

            resp = ServiceError("%s" % e)
            return resp(environ, start_response)


def make_app(satosa_config):
    try:
        default_logging_config = {
            "version": 1,
            "formatters": {
                "simple": {
                    "format": "[%(asctime)s] [%(levelname)s] [%(name)s.%(funcName)s] %(message)s"
                }
            },
            "handlers": {
                "stdout": {
                    "class": "logging.StreamHandler",
                    "stream": "ext://sys.stdout",
                    "level": "DEBUG",
                    "formatter": "simple",
                }
            },
            "loggers": {"satosa": {"level": "DEBUG"}},
            "root": {"level": "DEBUG", "handlers": ["stdout"]},
        }
        logging.config.dictConfig(satosa_config.get("LOGGING", default_logging_config))

        logger.info("Running SATOSA version {v}".format(v=satosa.__version__))

        res1 = WsgiApplication(satosa_config)
        res2 = CookiesSameSiteCompatMiddleware(res1, satosa_config)
        res3 = ToBytesMiddleware(res2)
        res = res3

        return res
    except Exception:
        logline = "Failed to create WSGI app."
        logger.exception(logline)
        raise



"""
OAuth backend for BitBucket
"""
import json
import logging
import requests

from oic.utils.authn.authn_context import UNSPECIFIED
from oic.oauth2.consumer import stateID

from satosa.backends.oauth import _OAuthBackend
from satosa.internal import AuthenticationInformation

logger = logging.getLogger(__name__)


class BitBucketBackend(_OAuthBackend):
    """BitBucket OAuth 2.0 backend"""

    logprefix = "BitBucket Backend:"

    def __init__(self, outgoing, internal_attributes, config, base_url, name):
        """BitBucket backend constructor
        :param outgoing: Callback should be called by the module after the
            authorization in the backend is done.
        :param internal_attributes: Mapping dictionary between SATOSA internal
            attribute names and the names returned by underlying IdP's/OP's as
            well as what attributes the calling SP's and RP's expects namevice.
        :param config: configuration parameters for the module.
        :param base_url: base url of the service
        :param name: name of the plugin
        :type outgoing:
            (satosa.context.Context, satosa.internal.InternalData) ->
            satosa.response.Response
        :type internal_attributes: dict[string, dict[str, str | list[str]]]
        :type config: dict[str, dict[str, str] | list[str] | str]
        :type base_url: str
        :type name: str
        """
        config.setdefault('response_type', 'code')
        config['verify_accesstoken_state'] = False
        super().__init__(outgoing, internal_attributes, config, base_url,
                         name, 'bitbucket', 'account_id')

    def get_request_args(self, get_state=stateID):
        request_args = super().get_request_args(get_state=get_state)

        client_id = self.config["client_config"]["client_id"]
        extra_args = {
            arg_name: arg_val
            for arg_name in ["auth_type", "scope"]
            for arg_val in [self.config.get(arg_name, [])]
            if arg_val
        }
        extra_args.update({"client_id": client_id})
        request_args.update(extra_args)
        return request_args

    def auth_info(self, request):
        return AuthenticationInformation(
            UNSPECIFIED, None,
            self.config['server_info']['authorization_endpoint'])

    def user_information(self, access_token):
        url = self.config['server_info']['user_endpoint']
        email_url = "{}/emails".format(url)
        headers = {'Authorization': 'Bearer {}'.format(access_token)}
        resp = requests.get(url, headers=headers)
        data = json.loads(resp.text)
        if 'email' in self.config['scope']:
            resp = requests.get(email_url, headers=headers)
            emails = json.loads(resp.text)
            data.update({
                'email': [e for e in [d.get('email')
                                      for d in emails.get('values')
                                      if d.get('is_primary')
                                      ]
                          ],
                'email_confirmed': [e for e in [d.get('email')
                                                for d in emails.get('values')
                                                if d.get('is_confirmed')
                                                ]
                                    ]
                })
        return data



from warnings import warn as _warn

from satosa.exception import SATOSABadContextError


class Context(object):
    """
    Holds methods for sharing proxy data through the current request
    """
    KEY_METADATA_STORE = 'metadata_store'
    KEY_TARGET_ENTITYID = 'target_entity_id'
    KEY_FORCE_AUTHN = 'force_authn'
    KEY_MEMORIZED_IDP = 'memorized_idp'
    KEY_REQUESTER_METADATA = 'requester_metadata'
    KEY_AUTHN_CONTEXT_CLASS_REF = 'authn_context_class_ref'
    KEY_TARGET_AUTHN_CONTEXT_CLASS_REF = 'target_authn_context_class_ref'

    def __init__(self):
        self._path = None
        self.request = None
        self.request_uri = None
        self.request_method = None
        self.qs_params = None
        self.server = None
        self.http_headers = None
        self.cookie = None
        self.request_authorization = None
        self.target_backend = None
        self.target_frontend = None
        self.target_micro_service = None
        # This dict is a data carrier between frontend and backend modules.
        self.internal_data = {}
        self.state = None

    @property
    def KEY_BACKEND_METADATA_STORE(self):
        msg = "'{old_key}' is deprecated; use '{new_key}' instead.".format(
            old_key="KEY_BACKEND_METADATA_STORE", new_key="KEY_METADATA_STORE"
        )
        _warn(msg, DeprecationWarning)
        return Context.KEY_METADATA_STORE

    @property
    def path(self):
        """
        Get the path

        :rtype: str

        :return: context path
        """
        return self._path

    @path.setter
    def path(self, p):
        """
        Inserts a path to the context.
        This path is striped by the base_url, so for example:
            A path BASE_URL/ENDPOINT_URL, would be inserted as only ENDPOINT_URL
            https://localhost:8092/sso/redirect -> sso/redirect

        :type p: str

        :param p: A path to an endpoint.
        :return: None
        """
        if not p:
            raise ValueError("path can't be set to None")
        elif p.startswith('/'):
            raise ValueError("path can't start with '/'")
        self._path = p

    def target_entity_id_from_path(self):
        target_entity_id = self.path.split("/")[1]
        return target_entity_id

    def decorate(self, key, value):
        """
        Add information to the context
        """

        self.internal_data[key] = value
        return self

    def get_decoration(self, key):
        """
        Retrieve information from the context
        """

        value = self.internal_data.get(key)
        return value



"""
The SATOSA main module
"""
import json
import logging
import uuid

from saml2.s_utils import UnknownSystemEntity

from satosa import util
from satosa.response import BadRequest
from satosa.response import NotFound
from satosa.response import Redirect
from .context import Context
from .exception import SATOSAAuthenticationError
from .exception import SATOSAAuthenticationFlowError
from .exception import SATOSABadRequestError
from .exception import SATOSAError
from .exception import SATOSAMissingStateError
from .exception import SATOSANoBoundEndpointError
from .exception import SATOSAUnknownError
from .exception import SATOSAStateError
from .plugin_loader import load_backends
from .plugin_loader import load_frontends
from .plugin_loader import load_request_microservices
from .plugin_loader import load_response_microservices
from .routing import ModuleRouter
from .state import State
from .state import cookie_to_state
from .state import state_to_cookie

import satosa.logging_util as lu


logger = logging.getLogger(__name__)

STATE_KEY = "SATOSA_BASE"


class SATOSABase(object):
    """
    Base class for a satosa proxy server.
    Does not contain any server parts.
    """

    def __init__(self, config):
        """
        Creates a satosa proxy base

        :type config: satosa.satosa_config.SATOSAConfig
        :param config: satosa proxy config
        """
        self.config = config

        logger.info("Loading backend modules...")
        backends = load_backends(self.config, self._auth_resp_callback_func,
                                 self.config["INTERNAL_ATTRIBUTES"])
        logger.info("Loading frontend modules...")
        frontends = load_frontends(self.config, self._auth_req_callback_func,
                                   self.config["INTERNAL_ATTRIBUTES"])

        self.response_micro_services = []
        self.request_micro_services = []
        logger.info("Loading micro services...")
        if "MICRO_SERVICES" in self.config:
            self.request_micro_services.extend(load_request_microservices(
                self.config.get("CUSTOM_PLUGIN_MODULE_PATHS"),
                self.config["MICRO_SERVICES"],
                self.config["INTERNAL_ATTRIBUTES"],
                self.config["BASE"]))
            self._link_micro_services(self.request_micro_services, self._auth_req_finish)

            self.response_micro_services.extend(
                load_response_microservices(self.config.get("CUSTOM_PLUGIN_MODULE_PATHS"),
                                            self.config["MICRO_SERVICES"],
                                            self.config["INTERNAL_ATTRIBUTES"],
                                            self.config["BASE"]))
            self._link_micro_services(self.response_micro_services, self._auth_resp_finish)

        self.module_router = ModuleRouter(frontends, backends,
                                          self.request_micro_services + self.response_micro_services)

    def _link_micro_services(self, micro_services, finisher):
        if not micro_services:
            return

        for i in range(len(micro_services) - 1):
            micro_services[i].next = micro_services[i + 1].process

        micro_services[-1].next = finisher

    def _auth_req_callback_func(self, context, internal_request):
        """
        This function is called by a frontend module when an authorization request has been
        processed.

        :type context: satosa.context.Context
        :type internal_request: satosa.internal.InternalData
        :rtype: satosa.response.Response

        :param context: The request context
        :param internal_request: request processed by the frontend

        :return: response
        """
        state = context.state
        state[STATE_KEY] = {"requester": internal_request.requester}

        msg = "Requesting provider: {}".format(internal_request.requester)
        logline = lu.LOG_FMT.format(id=lu.get_session_id(state), message=msg)
        logger.info(logline)

        if self.request_micro_services:
            return self.request_micro_services[0].process(context, internal_request)

        return self._auth_req_finish(context, internal_request)

    def _auth_req_finish(self, context, internal_request):
        backend = self.module_router.backend_routing(context)
        context.request = None
        return backend.start_auth(context, internal_request)

    def _auth_resp_finish(self, context, internal_response):
        user_id_to_attr = self.config["INTERNAL_ATTRIBUTES"].get("user_id_to_attr", None)
        if user_id_to_attr:
            internal_response.attributes[user_id_to_attr] = [internal_response.subject_id]

        # remove all session state unless CONTEXT_STATE_DELETE is False
        context.state.delete = self.config.get("CONTEXT_STATE_DELETE", True)
        context.request = None

        frontend = self.module_router.frontend_routing(context)
        return frontend.handle_authn_response(context, internal_response)

    def _auth_resp_callback_func(self, context, internal_response):
        """
        This function is called by a backend module when the authorization is
        complete.

        :type context: satosa.context.Context
        :type internal_response: satosa.internal.InternalData
        :rtype: satosa.response.Response

        :param context: The request context
        :param internal_response: The authentication response
        :return: response
        """

        context.request = None
        internal_response.requester = context.state[STATE_KEY]["requester"]

        # If configured construct the user id from attribute values.
        if "user_id_from_attrs" in self.config["INTERNAL_ATTRIBUTES"]:
            subject_id = [
                "".join(internal_response.attributes[attr]) for attr in
                self.config["INTERNAL_ATTRIBUTES"]["user_id_from_attrs"]
            ]
            internal_response.subject_id = "".join(subject_id)

        if self.response_micro_services:
            return self.response_micro_services[0].process(
                context, internal_response)

        return self._auth_resp_finish(context, internal_response)

    def _handle_satosa_authentication_error(self, error):
        """
        Sends a response to the requester about the error

        :type error: satosa.exception.SATOSAAuthenticationError
        :rtype: satosa.response.Response

        :param error: The exception
        :return: response
        """
        context = Context()
        context.state = error.state
        frontend = self.module_router.frontend_routing(context)
        return frontend.handle_backend_error(error)

    def _run_bound_endpoint(self, context, spec):
        """

        :type context: satosa.context.Context
        :type spec: ((satosa.context.Context, Any) -> satosa.response.Response, Any) |
        (satosa.context.Context) -> satosa.response.Response

        :param context: The request context
        :param spec: bound endpoint function
        :return: response
        """
        try:
            return spec(context)
        except SATOSAAuthenticationError as error:
            error.error_id = uuid.uuid4().urn
            state = json.dumps(error.state.state_dict, indent=4)
            msg = "ERROR_ID [{err_id}]\nSTATE:\n{state}".format(
                err_id=error.error_id, state=state
            )
            logline = lu.LOG_FMT.format(id=lu.get_session_id(context.state), message=msg)
            logger.error(logline, exc_info=True)
            return self._handle_satosa_authentication_error(error)

    def _load_state(self, context):
        """
        Load state from cookie to the context

        :type context: satosa.context.Context
        :param context: Session context
        """
        try:
            state = cookie_to_state(
                context.cookie,
                self.config["COOKIE_STATE_NAME"],
                self.config["STATE_ENCRYPTION_KEY"],
            )
        except SATOSAStateError:
            state = State()
        finally:
            context.state = state
            msg = f"Loaded state {state} from cookie {context.cookie}"
            logline = lu.LOG_FMT.format(id=lu.get_session_id(context.state), message=msg)
            logger.debug(logline)

    def _save_state(self, resp, context):
        """
        Saves a state from context to cookie

        :type resp: satosa.response.Response
        :type context: satosa.context.Context

        :param resp: The response
        :param context: Session context
        """

        cookie_name = self.config["COOKIE_STATE_NAME"]
        cookie = state_to_cookie(
            context.state,
            name=cookie_name,
            path="/",
            encryption_key=self.config["STATE_ENCRYPTION_KEY"],
            secure=self.config.get("COOKIE_SECURE"),
            httponly=self.config.get("COOKIE_HTTPONLY"),
            samesite=self.config.get("COOKIE_SAMESITE"),
            max_age=self.config.get("COOKIE_MAX_AGE"),
        )
        resp.headers = [
            (name, value)
            for (name, value) in resp.headers
            if name != "Set-Cookie"
            or not value.startswith(f"{cookie_name}=")
        ]
        resp.headers.append(tuple(cookie.output().split(": ", 1)))

    def run(self, context):
        """
        Runs the satosa proxy with the given context.

        :type context: satosa.context.Context
        :rtype: satosa.response.Response

        :param context: The request context
        :return: response
        """
        try:
            self._load_state(context)
            spec = self.module_router.endpoint_routing(context)
            resp = self._run_bound_endpoint(context, spec)
            self._save_state(resp, context)
        except SATOSABadRequestError as e:
            error_id = uuid.uuid4().urn
            msg = {
                "message": "Bad Request",
                "error": str(e),
                "error_id": error_id,
            }
            logline = lu.LOG_FMT.format(id=lu.get_session_id(context.state), message=msg)
            logger.error(logline)
            generic_error_url = self.config.get("ERROR_URL")
            if generic_error_url:
                redirect_url = f"{generic_error_url}?errorid={error_id}"
                return Redirect(generic_error_url)
            return BadRequest(e.error)
        except SATOSAMissingStateError as e:
            error_id = uuid.uuid4().urn
            msg = {
                "message": "Missing SATOSA State",
                "error": str(e),
                "error_id": error_id,
            }
            logline = lu.LOG_FMT.format(id=lu.get_session_id(context.state), message=msg)
            logger.error(logline)
            generic_error_url = self.config.get("ERROR_URL")
            if generic_error_url:
                redirect_url = f"{generic_error_url}?errorid={error_id}"
                return Redirect(generic_error_url)
            raise
        except SATOSAAuthenticationFlowError as e:
            error_id = uuid.uuid4().urn
            msg = {
                "message": "SATOSA Authentication Flow Error",
                "error": str(e),
                "error_id": error_id,
            }
            logline = lu.LOG_FMT.format(id=lu.get_session_id(context.state), message=msg)
            logger.error(logline)
            generic_error_url = self.config.get("ERROR_URL")
            if generic_error_url:
                redirect_url = f"{generic_error_url}?errorid={error_id}"
                return Redirect(generic_error_url)
            raise
        except SATOSANoBoundEndpointError as e:
            error_id = uuid.uuid4().urn
            msg = {
                "message": "URL-path is not bound to any endpoint function",
                "error": str(e),
                "error_id": error_id,
            }
            logline = lu.LOG_FMT.format(id=lu.get_session_id(context.state), message=msg)
            logger.error(logline)
            generic_error_url = self.config.get("ERROR_URL")
            if generic_error_url:
                redirect_url = f"{generic_error_url}?errorid={error_id}"
                return Redirect(generic_error_url)
            return NotFound("The Service or Identity Provider you requested could not be found.")
        except SATOSAError as e:
            error_id = uuid.uuid4().urn
            msg = {
                "message": "Uncaught SATOSA error",
                "error": str(e),
                "error_id": error_id,
            }
            logline = lu.LOG_FMT.format(id=lu.get_session_id(context.state), message=msg)
            logger.error(logline)
            generic_error_url = self.config.get("ERROR_URL")
            if generic_error_url:
                redirect_url = f"{generic_error_url}?errorid={error_id}"
                return Redirect(generic_error_url)
            raise
        except UnknownSystemEntity as e:
            error_id = uuid.uuid4().urn
            msg = {
                "message": "Configuration error: unknown system entity",
                "error": str(e),
                "error_id": error_id,
            }
            logline = lu.LOG_FMT.format(id=lu.get_session_id(context.state), message=msg)
            logger.error(logline)
            generic_error_url = self.config.get("ERROR_URL")
            if generic_error_url:
                redirect_url = f"{generic_error_url}?errorid={error_id}"
                return Redirect(generic_error_url)
            raise
        except Exception as e:
            error_id = uuid.uuid4().urn
            msg = {
                "message": "Uncaught exception",
                "error": str(e),
                "error_id": error_id,
            }
            logline = lu.LOG_FMT.format(id=lu.get_session_id(context.state), message=msg)
            logger.error(logline)
            generic_error_url = self.config.get("ERROR_URL")
            if generic_error_url:
                return Redirect(generic_error_url)
            raise SATOSAUnknownError("Unknown error") from e
        else:
            return resp


class SAMLBaseModule(object):
    KEY_ENTITYID_ENDPOINT = 'entityid_endpoint'
    KEY_ENABLE_METADATA_RELOAD = 'enable_metadata_reload'
    KEY_ATTRIBUTE_PROFILE = 'attribute_profile'
    KEY_ACR_MAPPING = 'acr_mapping'
    VALUE_ATTRIBUTE_PROFILE_DEFAULT = 'saml'

    def init_config(self, config):
        self.attribute_profile = config.get(
            self.KEY_ATTRIBUTE_PROFILE,
            self.VALUE_ATTRIBUTE_PROFILE_DEFAULT)
        self.acr_mapping = config.get(self.KEY_ACR_MAPPING)
        return config

    def expose_entityid_endpoint(self):
        value = self.config.get(self.KEY_ENTITYID_ENDPOINT, False)
        return bool(value)

    def enable_metadata_reload(self):
        """
        Check whether metadata reload has been enabled in config

        return: bool
        """
        value = self.config.get(self.KEY_ENABLE_METADATA_RELOAD, False)
        return bool(value)


class SAMLEIDASBaseModule(SAMLBaseModule):
    VALUE_ATTRIBUTE_PROFILE_DEFAULT = 'eidas'

    def init_config(self, config):
        config = super().init_config(config)

        spec_eidas = {
            'entityid_endpoint': True,
        }

        return util.check_set_dict_defaults(config, spec_eidas)



"""
Holds a base class for frontend modules used in the SATOSA proxy.
"""
from ..attribute_mapping import AttributeMapper


class FrontendModule(object):
    """
    Base class for a frontend module.
    """

    def __init__(self, auth_req_callback_func, internal_attributes, base_url, name):
        """
        :type auth_req_callback_func:
        (satosa.context.Context, satosa.internal.InternalData) -> satosa.response.Response
        :type internal_attributes: dict[str, dict[str, str | list[str]]]
        :type name: str

        :param auth_req_callback_func: Callback should be called by the module after the
        authorization response has been processed.
        :param name: name of the plugin
        """
        self.auth_req_callback_func = auth_req_callback_func
        self.internal_attributes = internal_attributes
        self.converter = AttributeMapper(internal_attributes)
        self.base_url = base_url
        self.name = name

    def handle_authn_response(self, context, internal_resp):
        """
        If an authorization has been successful in a backend, this function is called and is
        supposed to send an authorization response to the client.

        :type context: satosa.context.Context
        :type internal_resp: satosa.internal.InternalData
        :rtype satosa.response.Response

        :param context: The request context
        :param internal_resp: Attributes from the authorization
        :return response
        """
        raise NotImplementedError()

    def handle_backend_error(self, exception):
        """
        IF the backend gets an unexpected error, a suitable notice about the failure should be sent
        to the requester. This function is supposed to send a suitable error message to the
        requester.

        :type exception: satosa.exception.SATOSAError
        :rtype: satosa.response.Response

        :param exception: The raised exception
        :return: response
        """
        raise NotImplementedError()

    def register_endpoints(self, backend_names):
        """
        Register frontend functions to endpoint urls.

        Example of registering an endpoint:
            providers = ["Saml2IDP", "OIDCOP"]
            reg_endp = [
                ("^Saml2IDP/sso/redirect$", endpoint_function),
                ("^OIDCOP/sso/redirect$", endpoint_function),
            ]


        :type backend_names: list[str]
        :rtype List[Tuple[str, Callable[[satosa.context.Context, Any], satosa.response.Response]]]

        :param backend_names: Names of all all configured backends.
            All regexes produced for the frontends authentication endpoint must contain each backend name, e.g.:

                urls = []
                for name in backend_names:
                    urls.append("{}/authentication".format(name))

                urls.append("global_endpoint")
                return urls

        :return: A list with functions and args bound to a specific endpoint url,
                 [(regexp, function), ...]
        """
        raise NotImplementedError()



"""
Apple backend module.
"""
import logging
from .openid_connect import OpenIDConnectBackend, STATE_KEY
from oic.oauth2.message import Message
from oic.oic.message import AuthorizationResponse
import satosa.logging_util as lu
from ..exception import SATOSAAuthenticationError
import json
import requests


logger = logging.getLogger(__name__)


# https://developer.okta.com/blog/2019/06/04/what-the-heck-is-sign-in-with-apple
class AppleBackend(OpenIDConnectBackend):
    """Sign in with Apple backend"""

    def _get_tokens(self, authn_response, context):
        """
        :param authn_response: authentication response from OP
        :type authn_response: oic.oic.message.AuthorizationResponse
        :return: access token and ID Token claims
        :rtype: Tuple[Optional[str], Optional[Mapping[str, str]]]
        """
        if "code" in authn_response:
            # make token request
            # https://developer.apple.com/documentation/sign_in_with_apple/generate_and_validate_tokens
            args = {
                "client_id": self.client.client_id,
                "client_secret": self.client.client_secret,
                "code": authn_response["code"],
                "grant_type": "authorization_code",
                "redirect_uri": self.client.registration_response["redirect_uris"][0],
            }

            token_resp = requests.post(
                "https://appleid.apple.com/auth/token",
                data=args,
                headers={"Content-Type": "application/x-www-form-urlencoded"},
            ).json()

            logger.debug("apple response received")
            logger.debug(token_resp)

            self._check_error_response(token_resp, context)

            keyjar = self.client.keyjar
            id_token_claims = dict(
                Message().from_jwt(token_resp["id_token"], keyjar=keyjar)
            )

            return token_resp["access_token"], id_token_claims

        return authn_response.get("access_token"), authn_response.get("id_token")

    def response_endpoint(self, context, *args):
        """
        Handles the authentication response from the OP.
        :type context: satosa.context.Context
        :type args: Any
        :rtype: satosa.response.Response

        :param context: SATOSA context
        :param args: None
        :return:
        """
        backend_state = context.state[self.name]

        # Apple has no userinfo endpoint
        # but may send some user information via POST in the first request.
        #
        # References:
        # - https://developer.apple.com/documentation/sign_in_with_apple/sign_in_with_apple_rest_api/authenticating_users_with_sign_in_with_apple
        # - https://developer.apple.com/documentation/sign_in_with_apple/namei
        try:
            userdata = context.request.get("user", "{}")
            userinfo = json.loads(userdata)
        except json.JSONDecodeError:
            userinfo = {}

        authn_resp = self.client.parse_response(
            AuthorizationResponse, info=context.request, sformat="dict"
        )
        if backend_state[STATE_KEY] != authn_resp["state"]:
            msg = "Missing or invalid state in authn response for state: {}".format(
                backend_state
            )
            logline = lu.LOG_FMT.format(
                id=lu.get_session_id(context.state), message=msg
            )
            logger.debug(logline)
            raise SATOSAAuthenticationError(
                context.state, "Missing or invalid state in authn response"
            )

        self._check_error_response(authn_resp, context)
        access_token, id_token_claims = self._get_tokens(authn_resp, context)
        if not id_token_claims:
            id_token_claims = {}

        if not id_token_claims and not userinfo:
            msg = "No id_token or userinfo, nothing to do.."
            logline = lu.LOG_FMT.format(
                id=lu.get_session_id(context.state), message=msg
            )
            logger.error(logline)
            raise SATOSAAuthenticationError(context.state, "No user info available.")

        all_user_claims = dict(list(userinfo.items()) + list(id_token_claims.items()))

        # convert "string or Boolean" claims to actual booleans
        for bool_claim_name in ["email_verified", "is_private_email"]:
            if type(all_user_claims.get(bool_claim_name)) == str:
                all_user_claims[bool_claim_name] = (
                    True if all_user_claims[bool_claim_name] == "true" else False
                )

        msg = "UserInfo: {}".format(all_user_claims)
        logline = lu.LOG_FMT.format(id=lu.get_session_id(context.state), message=msg)
        logger.debug(logline)
        internal_resp = self._translate_response(
            all_user_claims, self.client.authorization_endpoint
        )
        return self.auth_callback_func(context, internal_resp)



import re
import logging

from .base import ResponseMicroService
from ..context import Context
from ..exception import SATOSAError

logger = logging.getLogger(__name__)

class AddStaticAttributes(ResponseMicroService):
    """
    Add static attributes to the responses.
    """

    def __init__(self, config, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.static_attributes = config["static_attributes"]

    def process(self, context, data):
        data.attributes.update(self.static_attributes)
        return super().process(context, data)


class FilterAttributeValues(ResponseMicroService):
    """
    Filter attribute values, only preserving those matching the given regex.
    """

    def __init__(self, config, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.attribute_filters = config["attribute_filters"]

    def process(self, context, data):
        # apply default filters
        provider_filters = self.attribute_filters.get("", {})
        target_provider = data.auth_info.issuer
        self._apply_requester_filters(data.attributes, provider_filters, data.requester, context, target_provider)

        # apply target provider specific filters
        provider_filters = self.attribute_filters.get(target_provider, {})
        self._apply_requester_filters(data.attributes, provider_filters, data.requester, context, target_provider)
        return super().process(context, data)

    def _apply_requester_filters(self, attributes, provider_filters, requester, context, target_provider):
        # apply default requester filters
        default_requester_filters = provider_filters.get("", {})
        self._apply_filters(attributes, default_requester_filters, context, target_provider)

        # apply requester specific filters
        requester_filters = provider_filters.get(requester, {})
        self._apply_filters(attributes, requester_filters, context, target_provider)

    def _apply_filters(self, attributes, attribute_filters, context, target_provider):
        for attribute_name, attribute_filters in attribute_filters.items():
            if type(attribute_filters) == str:
                # convert simple notation to filter list
                attribute_filters = {'regexp': attribute_filters}

            for filter_type, filter_value in attribute_filters.items():

                if filter_type == "regexp":
                    filter_func = re.compile(filter_value).search
                elif filter_type == "shibmdscope_match_scope":
                    mdstore = context.get_decoration(Context.KEY_METADATA_STORE)
                    md_scopes = list(mdstore.shibmd_scopes(target_provider,"idpsso_descriptor")) if mdstore else []
                    filter_func = lambda v: self._shibmdscope_match_scope(v, md_scopes)
                elif filter_type == "shibmdscope_match_value":
                    mdstore = context.get_decoration(Context.KEY_METADATA_STORE)
                    md_scopes = list(mdstore.shibmd_scopes(target_provider,"idpsso_descriptor")) if mdstore else []
                    filter_func = lambda v: self._shibmdscope_match_value(v, md_scopes)
                else:
                    raise SATOSAError("Unknown filter type")

                if attribute_name == "":  # default filter for all attributes
                    for attribute, values in attributes.items():
                        attributes[attribute] = list(filter(filter_func, attributes[attribute]))
                elif attribute_name in attributes:
                    attributes[attribute_name] = list(filter(filter_func, attributes[attribute_name]))

    def _shibmdscope_match_value(self, value, md_scopes):
        for md_scope in md_scopes:
            if not md_scope['regexp'] and md_scope['text'] == value:
                return True
            elif md_scope['regexp'] and re.fullmatch(md_scope['text'], value):
                return True
        return False

    def _shibmdscope_match_scope(self, value, md_scopes):
        split_value = value.split('@')
        if len(split_value) != 2:
            logger.info(f"Discarding invalid scoped value {value}")
            return False
        value_scope = split_value[1]
        return self._shibmdscope_match_value(value_scope, md_scopes)



"""
Exceptions for SATOSA
"""


class SATOSAError(Exception):
    """
    Base SATOSA exception
    """
    pass


class SATOSAConfigurationError(SATOSAError):
    """
    SATOSA configuration error
    """
    pass


class SATOSAStateError(SATOSAError):
    """
    SATOSA state error.
    """
    pass


class SATOSACriticalError(SATOSAError):
    """
    SATOSA critical error
    """
    pass


class SATOSAUnknownError(SATOSAError):
    """
    SATOSA unknown error
    """
    pass


class SATOSAAuthenticationError(SATOSAError):
    """
    SATOSA authentication error.
    """

    def __init__(self, state, message, *args, **kwargs):
        """
        :type state: satosa.state.State
        :param message: str
        :param args: any
        :param kwargs: any

        :param state: Satosa state
        :param message: A test message
        :param args: whatever
        :param kwargs: whatever
        """
        super().__init__(message, *args, **kwargs)
        self._message = "Authentication failed. Error id [{error_id}]"
        self.state = state.copy()
        self.error_id = 0

    @property
    def message(self):
        """
        :rtype: str
        :return: Exception message
        """
        return self._message.format(error_id=self.error_id)


class SATOSABasicError(SATOSAError):
    """
    eduTEAMS error
    """
    def __init__(self, error):
        self.error = error


class SATOSAMissingStateError(SATOSABasicError):
    """
    SATOSA Missing State error.

    This exception should be raised when SATOSA receives a request as part of
    an authentication flow and while the session state cookie is expected for
    that step, it is not included in the request
    """
    pass


class SATOSAAuthenticationFlowError(SATOSABasicError):
    """
    SATOSA Flow error.

    This exception should be raised when SATOSA receives a request that cannot
    be serviced because previous steps in the authentication flow for that session
    cannot be found
    """
    pass


class SATOSABadRequestError(SATOSABasicError):
    """
    SATOSA Bad Request error.

    This exception should be raised when we want to return an HTTP 400 Bad Request
    """
    pass


class SATOSABadContextError(SATOSAError):
    """
    Raise this exception if validating the Context and failing.
    """
    pass


class SATOSANoBoundEndpointError(SATOSAError):
    """
    Raised when a given url path is not bound to any endpoint function
    """
    pass



"""
This module contains all needed to keep a request state, without saving any information on the
server.
"""
import base64
import copy
import hashlib
import json
import logging
from collections import UserDict
from satosa.cookies import SimpleCookie
from uuid import uuid4

from lzma import LZMACompressor, LZMADecompressor

from Cryptodome import Random
from Cryptodome.Cipher import AES

import satosa.logging_util as lu
from satosa.exception import SATOSAStateError


logger = logging.getLogger(__name__)

_SESSION_ID_KEY = "SESSION_ID"


class State(UserDict):
    """
    This class holds a state attribute object. A state object must be able to be converted to
    a json string, otherwise will an exception be raised.
    """

    def __init__(self, urlstate_data=None, encryption_key=None):
        """
        If urlstate is empty a new empty state instance will be returned.

        If urlstate is not empty the constructor will rebuild the state attribute objects
        from the urlstate string.
        :type urlstate_data: str
        :type encryption_key: str
        :rtype: State

        :param encryption_key: The key to be used for encryption.
        :param urlstate_data: A string created by the method urlstate in this class.
        :return: An instance of this class.
        """
        self.delete = False

        urlstate_data = {} if urlstate_data is None else urlstate_data
        if urlstate_data and not encryption_key:
            raise ValueError("If an 'urlstate_data' is supplied 'encrypt_key' must be specified.")

        if urlstate_data:
            try:
                urlstate_data_bytes = urlstate_data.encode("utf-8")
                urlstate_data_b64decoded = base64.urlsafe_b64decode(urlstate_data_bytes)
                lzma = LZMADecompressor()
                urlstate_data_decompressed = lzma.decompress(urlstate_data_b64decoded)
                urlstate_data_decrypted = _AESCipher(encryption_key).decrypt(
                    urlstate_data_decompressed
                )
                lzma = LZMADecompressor()
                urlstate_data_decrypted_decompressed = lzma.decompress(urlstate_data_decrypted)
                urlstate_data_obj = json.loads(urlstate_data_decrypted_decompressed)
            except Exception as e:
                error_context = {
                    "message": "Failed to load state data. Reinitializing empty state.",
                    "reason": str(e),
                    "urlstate_data": urlstate_data,
                }
                logger.warning(error_context)
                urlstate_data = {}
            else:
                urlstate_data = urlstate_data_obj

        session_id = (
            urlstate_data[_SESSION_ID_KEY]
            if urlstate_data and _SESSION_ID_KEY in urlstate_data
            else uuid4().urn
        )
        urlstate_data[_SESSION_ID_KEY] = session_id

        super().__init__(urlstate_data)

    @property
    def session_id(self):
        return self.data.get(_SESSION_ID_KEY)

    def urlstate(self, encryption_key):
        """
        Will return a url safe representation of the state.

        :type encryption_key: Key used for encryption.
        :rtype: str

        :return: Url representation av of the state.
        """
        lzma = LZMACompressor()
        urlstate_data = json.dumps(self.data)
        urlstate_data = lzma.compress(urlstate_data.encode("UTF-8"))
        urlstate_data += lzma.flush()
        urlstate_data = _AESCipher(encryption_key).encrypt(urlstate_data)
        lzma = LZMACompressor()
        urlstate_data = lzma.compress(urlstate_data)
        urlstate_data += lzma.flush()
        urlstate_data = base64.urlsafe_b64encode(urlstate_data)
        return urlstate_data.decode("utf-8")

    def copy(self):
        """
        Returns a deepcopy of the state

        :rtype: satosa.state.State

        :return: A copy of the state
        """
        state_copy = State()
        state_copy.data = copy.deepcopy(self.data)
        return state_copy

    @property
    def state_dict(self):
        """
        :rtype: dict[str, any]
        :return: A copy of the state as dictionary.
        """
        return copy.deepcopy(self.data)


def state_to_cookie(
    state: State,
    *,
    name: str,
    path: str,
    encryption_key: str,
    secure: bool = None,
    httponly: bool = None,
    samesite: str = None,
    max_age: str = None,
) -> SimpleCookie:
    """
    Saves a state to a cookie

    :param state: the data to save
    :param name: identifier of the cookie
    :param path: path the cookie will be associated to
    :param encryption_key: the key to use to encrypt the state information
    :param secure: whether to include the cookie only when the request is transmitted
                   over a secure channel
    :param httponly: whether the cookie should only be accessed only by the server
    :param samesite: whether the cookie should only be sent with requests
                     initiated from the same registrable domain
    :param max_age: indicates the maximum lifetime of the cookie,
                    represented as the number of seconds until the cookie expires
    :return: A cookie object
    """
    cookie = SimpleCookie()
    cookie[name] = "" if state.delete else state.urlstate(encryption_key)
    cookie[name]["path"] = path
    cookie[name]["secure"] = secure if secure is not None else True
    cookie[name]["httponly"] = httponly if httponly is not None else ""
    cookie[name]["samesite"] = samesite if samesite is not None else "None"
    cookie[name]["max-age"] = (
        0
        if state.delete
        else max_age
        if max_age is not None
        else ""
    )

    msg = "Saved state in cookie {name} with properties {props}".format(
        name=name, props=list(cookie[name].items())
    )
    logline = lu.LOG_FMT.format(id=lu.get_session_id(state), message=msg)
    logger.debug(logline)

    return cookie


def cookie_to_state(cookie_str: str, name: str, encryption_key: str) -> State:
    """
    Loads a state from a cookie

    :type cookie_str: str
    :type name: str
    :type encryption_key: str
    :rtype: satosa.state.State

    :param cookie_str: string representation of cookie/s
    :param name: Name identifier of the cookie
    :param encryption_key: Key to encrypt the state information
    :return: A state
    """
    try:
        cookie = SimpleCookie(cookie_str)
        state = State(cookie[name].value, encryption_key)
    except KeyError as e:
        msg = f'No cookie named {name} in {cookie_str}'
        raise SATOSAStateError(msg) from e
    except ValueError as e:
        msg_tmpl = 'Failed to process {name} from {data}'
        msg = msg_tmpl.format(name=name, data=cookie_str)
        raise SATOSAStateError(msg) from e
    else:
        return state


class _AESCipher(object):
    """
    This class will perform AES encryption/decryption with a keylength of 256.

    @see: http://stackoverflow.com/questions/12524994/encrypt-decrypt-using-pycrypto-aes-256
    """

    def __init__(self, key):
        """
        Constructor

        :type key: str

        :param key: The key used for encryption and decryption. The longer key the better.
        """
        self.bs = 32
        self.key = hashlib.sha256(key.encode()).digest()

    def encrypt(self, raw):
        """
        Encryptes the parameter raw.

        :type raw: bytes
        :rtype: str

        :param: bytes to be encrypted.

        :return: A base 64 encoded string.
        """
        raw = self._pad(raw)
        iv = Random.new().read(AES.block_size)
        cipher = AES.new(self.key, AES.MODE_CBC, iv)
        return base64.urlsafe_b64encode(iv + cipher.encrypt(raw))

    def decrypt(self, enc):
        """
        Decryptes the parameter enc.

        :type enc: bytes
        :rtype: bytes

        :param: The value to be decrypted.
        :return: The decrypted value.
        """
        enc = base64.urlsafe_b64decode(enc)
        iv = enc[:AES.block_size]
        cipher = AES.new(self.key, AES.MODE_CBC, iv)
        return self._unpad(cipher.decrypt(enc[AES.block_size:]))

    def _pad(self, b):
        """
        Will padd the param to be of the correct length for the encryption alg.

        :type b: bytes
        :rtype: bytes
        """
        return b + (self.bs - len(b) % self.bs) * chr(self.bs - len(b) % self.bs).encode("UTF-8")

    @staticmethod
    def _unpad(b):
        """
        Removes the padding performed by the method _pad.

        :type b: bytes
        :rtype: bytes
        """
        return b[:-ord(b[len(b) - 1:])]
